; word count example of FASMLIB usage in MASM

; compile with:
;       ml /c /coff wc.asm
; link with:
;       link /subsystem:console /entry:start wc.obj "%FASMLIB%/bin/mscoff2.obj" kernel32.lib

; #########################################################################

      .586
      .model flat, stdcall
      option casemap:none    ; case sensitive

; #########################################################################

      %include @Environ(FASMLIB)\\include\\masm\\fasmlib.inc

; #########################################################################
  .data
	;prompts
	_prompt 	db "Please type name of file (max %d chars): ",0
	_result 	db "File contains %d words.",10,0

	;sample word delimiters
	_word_delims	db 9,10,13," ",".",",",":",0

	_err_format	db 10, "Error: %s!",10,0

	;file name buffer
	MAX_FILENAME=80+1
	filename	db MAX_FILENAME dup (?)

; #########################################################################

    .code

start:
	;initialize modules
	invoke	mem_init
	jc	error
	invoke	stream_init
	jc	error

	;write prompt
	push	MAX_FILENAME-1
	invoke	text_write_format, stream_stdout, offset _prompt
	jc	error

	;read file name
	invoke	text_read_line, stream_stdin, offset filename, MAX_FILENAME-1
	jc	error

	;EBX = open file as stream
	invoke	fstream_open, offset filename
	jc	error
	mov	ebx, eax

	;ECX = number of words
	xor	ecx, ecx

next_word:

	;skip all delimiters
	invoke	text_skip_chars, ebx, offset _word_delims

	;on ERR_HANDLE_EOF end loop
	.if (CARRY?)
	  cmp	  eax, ERR_HANDLE_EOF
	  je	  done
	  jmp	  error
	.endif

	;if we were on end of file, end loop
	jo	done

	;skip word
	invoke	text_skip_until, ebx, offset _word_delims

	;on ERR_HANDLE_EOF end loop
	.if (CARRY?)
	  cmp	  eax, ERR_HANDLE_EOF
	  je	  done
	  jmp	  error
	.endif

	;if we were on end of file, end loop
	jo	done

	;another word done
	inc	ecx

	jmp	next_word

done:
	;print result
	push	ecx
	invoke	text_write_format, stream_stdout, offset _result
	jc	error

	;close file
	invoke	stream_close, ebx
	jc	error

	;uninitialize modules
	invoke	stream_uninit
	jc	error
	invoke	mem_uninit
	jc	error

	;end program with exit code 0
	invoke	process_exit, 0



	;on error, EAX contains error code
error:
	;EAX = error message associated with error code in EAX
	invoke	err_text, eax

	;display error message
	push	eax
	invoke	text_write_format, stream_stderr, offset _err_format

	;unintialize modules
	invoke	stream_uninit
	invoke	mem_uninit

	;end program with exit code 1
	invoke	process_exit, 1

end start