;FASM STANDARD LIBRARY
;Conversion library

;b = unsigned byte     hex = zero-padded hexadecimal string    <floating formats>
;w = unsigned word     dec = decimal string
;d = unsigned dword    bin = binary string
;q = unsigned qword    oct = octal string
;ib = signed byte      dump = hex byte dump string
;iw = signed word      <float string>
;id = signed dword
;iq = signed qword


; proc conv.b2hex       dest, destlen, value
; proc conv.w2hex       dest, destlen, value
; proc conv.d2hex       dest, destlen, value
; proc conv.b2dec       dest, destlen, value
; proc conv.w2dec       dest, destlen, value
; proc conv.d2dec       dest, destlen, value
; proc conv.ib2hex      dest, destlen, value
; proc conv.iw2hex      dest, destlen, value
; proc conv.id2hex      dest, destlen, value
; proc conv.ib2dec      dest, destlen, value
; proc conv.iw2dec      dest, destlen, value
; proc conv.id2dec      dest, destlen, value

; proc conv.hex2b       string, strlen
; proc conv.hex2w       string, strlen
; proc conv.hex2d       string, strlen
; proc conv.dec2b       string, strlen
; proc conv.dec2w       string, strlen
; proc conv.dec2d       string, strlen

; proc extconv.b2hex    dest, destlen, value, padding, padlen, len
; proc extconv.w2hex    dest, destlen, value, padding, padlen, len
; proc extconv.d2hex    dest, destlen, value, padding, padlen, len
; proc extconv.b2dec    dest, destlen, value, padding, padlen, len
; proc extconv.w2dec    dest, destlen, value, padding, padlen, len
; proc extconv.d2dec    dest, destlen, value, padding, padlen, len
; proc extconv.ib2hex   dest, destlen, value, padding, padlen, plus, len
; proc extconv.ib2hex   dest, destlen, value, padding, padlen, plus, len
; proc extconv.id2hex   dest, destlen, value, padding, padlen, plus, len
; proc extconv.ib2dec   dest, destlen, value, padding, padlen, plus, len
; proc extconv.iw2dec   dest, destlen, value, padding, padlen, plus, len
; proc extconv.id2dec   dest, destlen, value, padding, padlen, plus, len

; proc extconv.hex2b    string, strlen, len
; proc extconv.hex2w    string, strlen, len
; proc extconv.hex2d    string, strlen, len
; proc extconv.dec2b    string, strlen, len
; proc extconv.dec2w    string, strlen, len
; proc extconv.dec2d    string, strlen, len


idata {
  conv.hextable db '0123456789ABCDEF'
}



;===========================================================================
; conv.b2hex
; desc: converts unsigned byte (8bit) value to hex string
; args: dest    - pointer to destination string buffer
;       destlen - size of destination buffer (3 is always enough)
;       value   - low order byte of this dword is the value to convert
; ret: CF=1 on error
; error: ERR_ZERO_SIZE   - destlen=0
;        ERR_BUFFER_FULL - destination buffer smaller than 3 bytes, only first
;                          destlen characters of result were stored
;===========================================================================
proc conv.b2hex dest, destlen, value
	movzx	eax, byte [value]
	libcall extconv.d2hex, [dest], [destlen], eax, 0, 0, 0
	ret
endp



;===========================================================================
; conv.w2hex
; desc: converts unsigned word (16bit) value to hex string. pads with zeroes
; args: dest    - pointer to destination string buffer
;       destlen - size of destination buffer (5 is always enough)
;       value   - value to convert, only low order word is used
; ret: CF=1 on error
; error: ERR_ZERO_SIZE   - destlen=0
;        ERR_BUFFER_FULL - destination buffer smaller than 5 bytes, only first
;                          destlen characters of result were stored
;===========================================================================
proc conv.w2hex dest, destlen, value
	movzx	eax, word [value]
	libcall extconv.d2hex, [dest], [destlen], eax, 0, 0, 0
	ret
endp



;===========================================================================
; conv.d2hex
; desc: converts unsigned dword (32bit) value to hex string. pads with zeroes
; args: dest    - pointer to destination string buffer
;       destlen - size of destination buffer (9 is always enough)
;       value   - value to convert
; ret: CF=1 on error
; error: ERR_ZERO_SIZE   - destlen=0
;        ERR_BUFFER_FULL - destination buffer smaller than 9 bytes, only first
;                          destlen characters of result were stored
;===========================================================================
proc conv.d2hex dest, destlen, value
	libcall extconv.d2hex, [dest], [destlen], [value], 0, 0, 0
	ret
endp



;=============================================================================
; conv.b2dec
; desc: converts unsigned byte (8bit) value to decimal string.
; args: dest    - pointer to destination string buffer
;       destlen - size of destination buffer (4 is always enough)
;       value   - value to convert
; ret: CF=1 on error
; error: ERR_ZERO_SIZE   - destlen=0.
;        ERR_BUFFER_FULL - destination buffer smaller than 4 bytes, only
;                          first destlen characters of result were stored.
;=============================================================================
proc conv.b2dec dest, destlen, value
	movzx	eax, byte [value]
	libcall extconv.d2dec, [dest], [destlen], eax, 0, 0, 0
	ret
endp



;=============================================================================
; conv.w2dec
; desc: converts unsigned word (16bit) value to decimal string.
; args: dest    - pointer to destination string buffer
;       destlen - size of destination buffer (7 is always enough)
;       value   - value to convert
; ret: CF=1 on error
; error: ERR_ZERO_SIZE   - destlen=0.
;        ERR_BUFFER_FULL - destination buffer smaller than 7 bytes, only
;                          first destlen characters of result were stored.
;=============================================================================
proc conv.w2dec dest, destlen, value
	movzx	eax, word [value]
	libcall extconv.d2dec, [dest], [destlen], eax, 0, 0, 0
	ret
endp



;=============================================================================
; conv.d2dec
; desc: converts unsigned dword (32bit) value to decimal string.
; args: dest    - pointer to destination string buffer
;       destlen - size of destination buffer (11 is always enough)
;       value   - value to convert
; ret: CF=1 on error
; error: ERR_ZERO_SIZE   - destlen=0.
;        ERR_BUFFER_FULL - destination buffer smaller than 11 bytes, only
;                          first destlen characters of result were stored.
;=============================================================================
proc conv.d2dec dest, destlen, value
	libcall extconv.d2dec, [dest], [destlen], [value], 0, 0, 0
	ret
endp



;============================================================================
; conv.ib2hex
; desc: converts signed byte (8bit) value to hex string.
; args: dest    - pointer to destination string buffer
;       destlen - size of destination buffer
;       value   - low order byte of this dword is value to convert
; ret: CF=1 on error
; error: ERR_ZERO_SIZE    - destlen=0. len set to 0
;        ERR_BUFFER_FULL  - destination buffer small. only first destlen
;                           characters of result were stored. len=destlen.
;============================================================================
proc conv.ib2hex dest, destlen, value
	movsx	eax, byte [value]
	libcall extconv.id2hex, [dest], [destlen], eax, 0, 0, 0, 0
	ret
endp



;============================================================================
; conv.iw2hex
; desc: converts signed word (16bit) value to hex string.
; args: dest    - pointer to destination string buffer
;       destlen - size of destination buffer
;       value   - low order byte of this dword is value to convert
; ret: CF=1 on error
; error: ERR_ZERO_SIZE    - destlen=0. len set to 0
;        ERR_BUFFER_FULL  - destination buffer small. only first destlen
;                           characters of result were stored. len=destlen.
;============================================================================
proc conv.iw2hex dest, destlen, value
	movsx	eax, word [value]
	libcall extconv.id2hex, [dest], [destlen], eax, 0, 0, 0, 0
	ret
endp


;============================================================================
; conv.id2hex
; desc: converts signed dword (32bit) value to hex string.
; args: dest    - pointer to destination string buffer
;       destlen - size of destination buffer
;       value   - low order byte of this dword is value to convert
; ret: CF=1 on error
; error: ERR_ZERO_SIZE    - destlen=0. len set to 0
;        ERR_BUFFER_FULL  - destination buffer small. only first destlen
;                           characters of result were stored. len=destlen.
;============================================================================
proc conv.id2hex dest, destlen, value
	libcall extconv.id2hex, [dest], [destlen], [value], 0, 0, 0, 0
	ret
endp



;============================================================================
; conv.ib2dec
; desc: converts signed byte (8bit) value to decimal string.
; args: dest    - pointer to destination string buffer
;       destlen - size of destination buffer
;       value   - low order byte of this dword is value to convert
; ret: CF=1 on error
; error: ERR_ZERO_SIZE    - destlen=0. len set to 0
;        ERR_BUFFER_FULL  - destination buffer small. only first destlen
;                           characters of result were stored. len=destlen.
;============================================================================
proc conv.ib2dec dest, destlen, value
	movsx	eax, byte [value]
	libcall extconv.id2dec, [dest], [destlen], eax, 0, 0, 0, 0
	ret
endp



;============================================================================
; conv.iw2dec
; desc: converts signed word (16bit) value to decimal string.
; args: dest    - pointer to destination string buffer
;       destlen - size of destination buffer
;       value   - low order byte of this dword is value to convert
; ret: CF=1 on error
; error: ERR_ZERO_SIZE    - destlen=0. len set to 0
;        ERR_BUFFER_FULL  - destination buffer small. only first destlen
;                           characters of result were stored. len=destlen.
;============================================================================
proc conv.iw2dec dest, destlen, value
	movsx	eax, word [value]
	libcall extconv.id2dec, [dest], [destlen], eax, 0, 0, 0, 0
	ret
endp



;============================================================================
; conv.id2dec
; desc: converts signed dword (32bit) value to decimal string.
; args: dest    - pointer to destination string buffer
;       destlen - size of destination buffer
;       value   - low order byte of this dword is value to convert
; ret: CF=1 on error
; error: ERR_ZERO_SIZE    - destlen=0. len set to 0
;        ERR_BUFFER_FULL  - destination buffer small. only first destlen
;                           characters of result were stored. len=destlen.
;============================================================================
proc conv.id2dec dest, destlen, value
	libcall extconv.id2dec, [dest], [destlen], [value], 0, 0, 0, 0
	ret
endp



;=============================================================================
; conv.hex2b
; desc: converts hex string (case insensitive) to unsigned byte (8bit) value
; args: string  - pointer to string which we are converting
;       strlen  - size of string buffer
; ret: CF=1 on error, otherwise
;      EAX = value (AL zero extended to EAX)
; note: converts hex digits until reaches non-digit, or value overflows
; error: ERR_ZERO_SIZE        - strlen=0
;        ERR_OUT_OF_RANGE     - value is > 0FFh
;        ERR_NOT_HEXADECIMAL  - first character is not hex digit
;=============================================================================
proc conv.hex2b string, strlen
	libcall extconv.hex2d_size, [string], [strlen], 0F0h, 0
	ret
endp



;=============================================================================
; conv.hex2w
; desc: converts hex string (case insensitive) to unsigned word (16bit) value
; args: string  - pointer to string which we are converting
;       strlen  - size of string buffer
; ret: CF=1 on error, otherwise
;      EAX = value (AX zero extended to EAX)
; note: converts hex digits until reaches non-digit, or value overflows
; error: ERR_ZERO_SIZE        - strlen=0
;        ERR_OUT_OF_RANGE     - value is > 0FFFFh
;        ERR_NOT_HEXADECIMAL  - first character is not hex digit
;=============================================================================
proc conv.hex2w string, strlen
	libcall extconv.hex2d_size, [string], [strlen], 0F000h, 0
	ret
endp



;=============================================================================
; conv.hex2d
; desc: converts hex string (case insensitive) to unsigned dword (32bit) value
; args: string  - pointer to string which we are converting
;       strlen  - size of string buffer
; ret: CF=1 on error, otherwise
;      EAX = value
; note: converts hex digits until reaches non-digit, or value overflows
; error: ERR_ZERO_SIZE        - strlen=0
;        ERR_OUT_OF_RANGE     - value is > 0FFFFFFFFh
;        ERR_NOT_HEXADECIMAL  - first character is not hex digit
;=============================================================================
proc conv.hex2d string, strlen
	libcall extconv.hex2d_size, [string], [strlen], 0F0000000h, 0
	ret
endp



;=============================================================================
; conv.dec2b
; desc: converts decimal string to unsigned byte (8bit) value
; args: string  - pointer to string which we are converting
;       strlen  - size of string buffer
; ret: CF=1 on error, otherwise
;      EAX = value (AL zero extended to dword)
; note: converts digits until reaches non-digit, or value overflows
; error: ERR_ZERO_SIZE    - strlen=0
;        ERR_OUT_OF_RANGE - value is > 255
;        ERR_NOT_DECIMAL  - first character is not decimal digit.
;=============================================================================
proc conv.dec2b string, strlen
	libcall extconv.dec2d_size, [string], [strlen], 255, 0
	ret
endp



;=============================================================================
; conv.dec2w
; desc: converts decimal string to unsigned word (16bit) value
; args: string  - pointer to string which we are converting
;       strlen  - size of string buffer
; ret: CF=1 on error, otherwise
;      EAX = value (AX zero extended to dword)
; note: converts hex digits until reaches non-digit, or value overflows
; error: ERR_ZERO_SIZE    - strlen=0
;        ERR_OUT_OF_RANGE - value is > 65535
;        ERR_NOT_DECIMAL  - first character is not decimal digit.
;=============================================================================
proc conv.dec2w string, strlen
	libcall extconv.dec2d_size, [string], [strlen], 65535, 0
	ret
endp



;============================================================================
; conv.dec2d
; desc: converts decimal string to unsigned dword (32bit) value
; args: string  - pointer to string which we are converting
;       strlen  - size of string buffer
; ret: CF=1 on error, otherwise
;      EAX = value
; note: converts hex digits until reaches non-digit, or value overflows
; error: ERR_ZERO_SIZE    - strlen=0
;        ERR_OUT_OF_RANGE - value is > 4294967295
;        ERR_NOT_DECIMAL  - first character is not decimal digit.
;============================================================================
proc conv.dec2d string, strlen
	libcall extconv.dec2d_size, [string], [strlen], 4294967295, 0
	ret
endp



;============================================================================
; extconv.b2hex
; desc: converts unsigned byte (8bit) value to hex string.
; args: dest    - pointer to destination string buffer
;       destlen - size of destination buffer
;       value   - low order byte of this dword is value to convert
;       padding - type of padding (0=none, 1=spaces, 2=zeroes)
;       padlen  - minimal length of number. ignored if padding=0.
;       len     - address of dword variable, which will on return hold number
;                 of characters of converted number. ignored if 0.
; ret: CF=1 on error
; error: ERR_ZERO_SIZE    - destlen=0. len set to 0
;        ERR_INVALID_MODE - bad padding type. len set to 0
;        ERR_BUFFER_FULL  - destination buffer small. only first destlen
;                           characters of result were stored. len=destlen.
;============================================================================
proc extconv.b2hex dest, destlen, value, padding, padlen, len
	movzx	eax, byte [value]
	libcall extconv.d2hex, [dest], [destlen], eax, [padding], [padlen], [len]
	ret
endp



;============================================================================
; extconv.w2hex
; desc: converts unsigned word (16bit) value to hex string.
; args: dest    - pointer to destination string buffer
;       destlen - size of destination buffer
;       value   - low order word of this dword is value to convert
;       padding - type of padding (0=none, 1=spaces, 2=zeroes)
;       padlen  - minimal length of number. ignored if padding=0.
;       len     - address of dword variable, which will on return hold number
;                 of characters of converted number. ignored if 0.
; ret: CF=1 on error
; error: ERR_ZERO_SIZE    - destlen=0. len set to 0
;        ERR_INVALID_MODE - bad padding type. len set to 0
;        ERR_BUFFER_FULL  - destination buffer small. only first destlen
;                           characters of result were stored. len=destlen.
;============================================================================
proc extconv.w2hex dest, destlen, value, padding, padlen, len
	movzx	eax, word [value]
	libcall extconv.d2hex, [dest], [destlen], eax, [padding], [padlen], [len]
	ret
endp



;============================================================================
; extconv.d2hex
; desc: converts unsigned dword (32bit) value to hex string.
; args: dest    - pointer to destination string buffer
;       destlen - size of destination buffer
;       value   - value to convert
;       padding - type of padding (0=none, 1=spaces, 2=zeroes)
;       padlen  - minimal length of number. ignored if padding=0.
;       len     - address of dword variable, which will on return hold number
;                 of characters of converted number. ignored if 0.
; ret: CF=1 on error
; error: ERR_ZERO_SIZE    - destlen=0. len set to 0
;        ERR_INVALID_MODE - bad padding type. len set to 0
;        ERR_BUFFER_FULL  - destination buffer small. only first destlen
;                           characters of result were stored. len=destlen.
;============================================================================
proc extconv.d2hex dest, destlen, value, padding, padlen, len
	push	ebx ecx edx edi

	cmp	[destlen], 0
	je	.error_destlen_zero
	cmp	[padding], 2
	ja	.error_padding_type

	mov	edi, [dest]
	mov	ebx, [destlen]

	;convert value (push hex digits to stack as dwords)
	xor	edx, edx
	xor	ecx, ecx		;ecx = number of chars
	mov	eax, [value]		;eax = value
.push_digit:
	mov	dl, al
	and	dl, 0Fh
	push	edx
	shr	eax, 4
	inc	ecx
	cmp	eax, 0
	jnz	.push_digit

	;get padding type
	cmp	[padding], 0
	je	.padded
	mov	dl, ' '
	cmp	[padding], 1
	je	@f
	mov	dl, '0'
	@@:

	;pad up to padlen
	mov	eax, [padlen]
.pad:	cmp	eax, ecx
	jbe	.padded
	mov	byte [edi], dl
	inc	edi
	dec	ebx
	jz	.error_buffer_full
	dec	eax
	jmp	.pad
.padded:

	;pop values from stack and store them
	;EDI = dest, EBX=destlen
.store_digit:
	pop	eax
	mov	al, [conv.hextable + eax]
	mov	[edi], al
	inc	edi
	dec	ebx
	jz	.error_buffer_full
	dec	ecx
	jnz	.store_digit

	;store ending zero
	mov	byte [edi], 0

	;save len
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	sub	edi, [dest]
	mov	[eax], edi
	@@:

	clc

.r:	pop	edi edx ecx ebx
	ret
.rc:	stc
	jmp	.r

.error_buffer_full:
	;restore stack
	lea	esp, [esp + 4*(ecx-1)]

	;set len to destlen
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	push	[destlen]
	pop	dword [eax]
	@@:

	mov	eax, ERR_BUFFER_FULL
	jmp	.rc

.error_destlen_zero:
	;set len to 0
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], 0
	@@:

	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_padding_type:
	;set len to 0
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], 0
	@@:

	mov	eax, ERR_INVALID_MODE
	jmp	.rc
endp



;=============================================================================
; extconv.b2dec
; desc: converts unsigned byte (8bit) value to decimal string.
; args: dest    - pointer to destination string buffer
;       destlen - size of destination buffer
;       value   - low order byte of this dword is value to convert
;       padding - type of padding (0=none, 1=spaces, 2=zeroes)
;       padlen  - minimal length of number. ignored if padding=0.
;       len     - address of dword variable, which will on return hold number
;                 of characters of converted number. ignored if 0.
; ret: CF=1 on error
; error: ERR_ZERO_SIZE    - destlen=0. len set to 0
;        ERR_INVALID_MODE - bad padding type. len set to 0
;        ERR_BUFFER_FULL  - destination buffer small. only first destlen
;                           characters of result were stored. len=destlen.
;=============================================================================
proc extconv.b2dec dest, destlen, value, padding, padlen, len
	movzx	eax, byte [value]
	libcall extconv.d2dec, [dest], [destlen], eax, [padding], [padlen], [len]
	ret
endp



;=============================================================================
; extconv.w2dec
; desc: converts unsigned word (16bit) value to decimal string.
; args: dest    - pointer to destination string buffer
;       destlen - size of destination buffer
;       value   - low order word of this dword is value to convert
;       padding - type of padding (0=none, 1=spaces, 2=zeroes)
;       padlen  - minimal length of number. ignored if padding=0.
;       len     - address of dword variable, which will on return hold number
;                 of characters of converted number. ignored if 0.
; ret: CF=1 on error
; error: ERR_ZERO_SIZE    - destlen=0. len set to 0
;        ERR_INVALID_MODE - bad padding type. len set to 0
;        ERR_BUFFER_FULL  - destination buffer small. only first destlen
;                           characters of result were stored. len=destlen.
;=============================================================================
proc extconv.w2dec dest, destlen, value, padding, padlen, len
	movzx	eax, word [value]
	libcall extconv.d2dec, [dest], [destlen], eax, [padding], [padlen], [len]
	ret
endp



;=============================================================================
; extconv.d2dec
; desc: converts unsigned dword (32bit) value to decimal string.
; args: dest    - pointer to destination string buffer
;       destlen - size of destination buffer
;       value   - value to convert
;       padding - type of padding (0=none, 1=spaces, 2=zeroes)
;       padlen  - minimal length of number. ignored if padding=0.
;       len     - address of dword variable, which will on return hold number
;                 of characters of converted number. ignored if 0.
; ret: CF=1 on error
; error: ERR_ZERO_SIZE    - destlen=0. len set to 0
;        ERR_INVALID_MODE - bad padding type. len set to 0
;        ERR_BUFFER_FULL  - destination buffer small. only first destlen
;                           characters of result were stored. len=destlen.
;=============================================================================
proc extconv.d2dec dest, destlen, value, padding, padlen, len
	push	ebx ecx edx edi

	cmp	[destlen], 0
	je	untaken .error_destlen_zero
	cmp	[padding], 2
	ja	untaken .error_padding_type

	mov	edi, [dest]

	;convert value (push remainders to stack as dwords)
	xor	ecx, ecx		;ecx = number of chars
	mov	eax, [value]		;eax = value
	mov	ebx, 10
.push_digit:
	xor	edx, edx
	div	ebx
	push	edx
	inc	ecx
	cmp	eax, 0
	jnz	taken .push_digit


	mov	ebx, [destlen]

	;get padding type
	cmp	[padding], 0
	je	.padded
	mov	dl, ' '
	cmp	[padding], 1
	je	@f
	mov	dl, '0'
	@@:

	;pad up to padlen
	mov	eax, [padlen]
.pad:	cmp	eax, ecx
	jbe	.padded
	mov	byte [edi], dl
	inc	edi
	dec	ebx
	jz	untaken .error_buffer_full
	dec	eax
	jmp	.pad
.padded:

	;pop values from stack and store them
	;EDI = dest, EBX=destlen
.store_digit:
	pop	eax
	mov	al, [conv.hextable + eax]
	mov	[edi], al
	inc	edi
	dec	ebx
	jz	.error_buffer_full
	dec	ecx
	jnz	taken .store_digit

	;store ending zero
	mov	byte [edi], 0

	;save len
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	sub	edi, [dest]
	mov	[eax], edi
	@@:

	clc

.r:	pop	edi edx ecx ebx
	ret
.rc:	stc
	jmp	.r

.error_buffer_full:
	;restore stack
	lea	esp, [esp + 4*(ecx-1)]

	;set len to destlen
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	push	[destlen]
	pop	dword [eax]
	@@:

	mov	eax, ERR_BUFFER_FULL
	jmp	.rc

.error_destlen_zero:
	;set len to 0
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], 0
	@@:

	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_padding_type:
	;set len to 0
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], 0
	@@:

	mov	eax, ERR_INVALID_MODE
	jmp	.rc
endp



;============================================================================
; extconv.ib2hex
; desc: converts signed byte (8bit) value to hex string.
; args: dest    - pointer to destination string buffer
;       destlen - size of destination buffer
;       value   - low order byte of this dword is value to convert
;       padding - type of padding (0=none, 1=spaces, 2=zeroes)
;       padlen  - minimal length of number. ignored if padding=0.
;       plus    - 0=no "+", 1="+" for positive numbers
;                 2=" " for positive, only useful when padding=2
;       len     - address of dword variable, which will on return hold number
;                 of characters of converted number. ignored if 0.
; ret: CF=1 on error
; error: ERR_ZERO_SIZE    - destlen=0. len set to 0
;        ERR_INVALID_MODE - bad padding or plus value. len set to 0
;        ERR_BUFFER_FULL  - destination buffer small. only first destlen
;                           characters of result were stored. len=destlen.
;============================================================================
proc extconv.ib2hex dest, destlen, value, padding, padlen, plus, len
	movsx	eax, byte [value]
	libcall extconv.id2hex, [dest], [destlen], eax, [padding], [padlen], [plus], [len]
	ret
endp



;============================================================================
; extconv.iw2hex
; desc: converts signed word (16bit) value to hex string.
; args: dest    - pointer to destination string buffer
;       destlen - size of destination buffer
;       value   - low order word of this dword is value to convert
;       padding - type of padding (0=none, 1=spaces, 2=zeroes)
;       padlen  - minimal length of number. ignored if padding=0.
;       plus    - 0=no "+", 1="+" for positive numbers
;                 2=" " for positive, only useful when padding=2
;       len     - address of dword variable, which will on return hold number
;                 of characters of converted number. ignored if 0.
; ret: CF=1 on error
; error: ERR_ZERO_SIZE    - destlen=0. len set to 0
;        ERR_INVALID_MODE - bad padding or plus value. len set to 0
;        ERR_BUFFER_FULL  - destination buffer small. only first destlen
;                           characters of result were stored. len=destlen.
;============================================================================
proc extconv.iw2hex dest, destlen, value, padding, padlen, plus, len
	movsx	eax, word [value]
	libcall extconv.id2hex, [dest], [destlen], eax, [padding], [padlen], [plus], [len]
	ret
endp



;============================================================================
; extconv.id2hex
; desc: converts signed dword (32bit) value to hex string.
; args: dest    - pointer to destination string buffer
;       destlen - size of destination buffer
;       value   - value to convert
;       padding - type of padding (0=none, 1=spaces, 2=zeroes)
;       padlen  - minimal length of number. ignored if padding=0.
;       plus    - 0=no plus sign, 1="+" for positive numbers
;                 2=" " for positive numbers (only useful when padding=2)
;       len     - address of dword variable, which will on return hold number
;                 of characters of converted number. ignored if 0.
; ret: CF=1 on error
; error: ERR_ZERO_SIZE    - destlen=0. len set to 0
;        ERR_INVALID_MODE - bad padding or plus value. len set to 0
;        ERR_BUFFER_FULL  - destination buffer small. only first destlen
;                           characters of result were stored. len=destlen.
;============================================================================
proc extconv.id2hex dest, destlen, value, padding, padlen, plus, len
	push	ebx ecx edx edi esi

	cmp	[destlen], 0
	je	untaken .error_destlen_zero
	cmp	[padding], 2
	ja	untaken .error_padding_type
	cmp	[plus], 2
	ja	untaken .error_plus

	mov	edi, [dest]
	mov	ebx, [destlen]

	;EAX = abs(value)
	mov	eax, [value]		;eax = value
	cmp	eax, 0
	jge	@f
	neg	eax
	@@:

	;convert value (push hex digits to stack as dwords)
	;ECX = number of digits on stack
	xor	edx, edx
	xor	ecx, ecx		;ecx = number of chars
.push_digit:
	mov	dl, al
	and	dl, 0Fh
	push	edx
	shr	eax, 4
	inc	ecx
	cmp	eax, 0
	jnz	taken .push_digit

.sign1:

	;get padding type
	cmp	[padding], 0
	je	.padded
	mov	dl, ' '
	cmp	[padding], 1
	je	@f
	mov	dl, '0'
	@@:

	;if padding type is "0", we must handle sign before padding
	cmp	dl, '0'
	jne	.do_padding

	;if value is negative, print '-'
	cmp	[value], 0
	jge	@f
	dec	ebx
	jz	.error_buffer_full
	mov	byte [edi], '-'
	inc	edi
	jmp	.do_padding
	@@:

	;value is positive, plus=0, don't print anything
	cmp	[plus], 0
	je	.do_padding

	;value is positive, print ' ' if plus=2
	cmp	[plus], 2
	jne	@f
	dec	ebx
	jz	.error_buffer_full
	mov	byte [edi], ' '
	inc	edi
	jmp	.do_padding
	@@:

	;value is positive, print '+' if plus=1
	cmp	[plus], 1
	jne	@f
	dec	ebx
	jz	.error_buffer_full
	mov	byte [edi], '+'
	inc	edi
	jmp	.do_padding
	@@:

	;do padding
	;EDI = dest, EBX=destlen, ECX=number of digits, DL=padding char
.do_padding:

	;padding disabled?
	cmp	[padding],0
	je	.padded

	;ESI = number of characters to pad
	mov	esi, ecx

	;if padding type=2, sign doesn't count into padding
	cmp	[padding],2
	je	.remove_sign_from_padding_done

	;if value<0 or plus!=0, increase ESI by one
	cmp	[value], 0
	jl	@f
	cmp	[plus], 0
	ja	@f
	jmp	.remove_sign_from_padding_done
	@@:
	inc	esi
.remove_sign_from_padding_done:

	;pad ESI up to padlen
	mov	eax, [padlen]
.pad:	cmp	eax, esi
	jbe	untaken .padded
	mov	byte [edi], dl
	inc	edi
	dec	ebx
	jz	untaken .error_buffer_full
	dec	eax
	jmp	.pad
.padded:

.sign2:
	;handle sign, if padding type is other than "0"
	cmp	[padding], 2
	je	.sign_done2

	;if value is negative, print '-'
	cmp	[value], 0
	jge	@f
	dec	ebx
	jz	.error_buffer_full
	mov	byte [edi], '-'
	inc	edi
	jmp	.sign_done2
	@@:

	;value is positive, print '+' if plus=1
	cmp	[plus], 1
	jne	.sign_done2
	dec	ebx
	jz	.error_buffer_full
	mov	byte [edi], '+'
	inc	edi

.sign_done2:

	;pop values from stack and store them
	;EDI = dest, EBX=destlen, ECX=number of digits on stack
.store_digit:
	pop	eax
	mov	al, [conv.hextable + eax]
	mov	[edi], al
	inc	edi
	dec	ebx
	jz	untaken .error_buffer_full
	dec	ecx
	jnz	taken .store_digit

	;store ending zero
	mov	byte [edi], 0

	;save len
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	sub	edi, [dest]
	mov	[eax], edi
	@@:

	clc

.r:	pop	esi edi edx ecx ebx
	ret
.rc:	stc
	jmp	.r

.error_buffer_full:
	;restore stack
	lea	esp, [esp + 4*(ecx-1)]

	;set len to destlen
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	push	[destlen]
	pop	dword [eax]
	@@:

	mov	eax, ERR_BUFFER_FULL
	jmp	.rc

.error_destlen_zero:
	;set len to 0
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], 0
	@@:

	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_plus:
.error_padding_type:
	;set len to 0
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], 0
	@@:

	mov	eax, ERR_INVALID_MODE
	jmp	.rc
endp



;=============================================================================
; extconv.ib2dec
; desc: converts signed byte (8bit) value to decimal string.
; args: dest    - pointer to destination string buffer
;       destlen - size of destination buffer
;       value   - low order byte of this dword is value to convert
;       padding - type of padding (0=none, 1=spaces, 2=zeroes)
;       padlen  - minimal length of number. ignored if padding=0.
;       plus    - 0=no "+", 1="+" for positive numbers
;                 2=" " for positive, only useful when padding=2
;       len     - address of dword variable, which will on return hold number
;                 of characters of converted number. ignored if 0.
; ret: CF=1 on error
; error: ERR_ZERO_SIZE    - destlen=0. len set to 0
;        ERR_INVALID_MODE - bad padding type. len set to 0
;        ERR_BUFFER_FULL  - destination buffer small. only first destlen
;                           characters of result were stored. len=destlen.
;=============================================================================
proc extconv.ib2dec dest, destlen, value, padding, padlen, plus, len
	movsx	eax, byte [value]
	libcall extconv.id2dec, [dest], [destlen], eax, [padding], [padlen], [plus], [len]
	ret
endp



;=============================================================================
; extconv.iw2dec
; desc: converts signed word (16bit) value to decimal string.
; args: dest    - pointer to destination string buffer
;       destlen - size of destination buffer
;       value   - low order word of this dword is value to convert
;       padding - type of padding (0=none, 1=spaces, 2=zeroes)
;       padlen  - minimal length of number. ignored if padding=0.
;       plus    - 0=no "+", 1="+" for positive numbers
;                 2=" " for positive, only useful when padding=2
;       len     - address of dword variable, which will on return hold number
;                 of characters of converted number. ignored if 0.
; ret: CF=1 on error
; error: ERR_ZERO_SIZE    - destlen=0. len set to 0
;        ERR_INVALID_MODE - bad padding type. len set to 0
;        ERR_BUFFER_FULL  - destination buffer small. only first destlen
;                           characters of result were stored. len=destlen.
;=============================================================================
proc extconv.iw2dec dest, destlen, value, padding, padlen, plus, len
	movsx	eax, word [value]
	libcall extconv.id2dec, [dest], [destlen], eax, [padding], [padlen], [plus], [len]
	ret
endp



;=============================================================================
; extconv.id2dec
; desc: converts signed dword (32bit) value to decimal string.
; args: dest    - pointer to destination string buffer
;       destlen - size of destination buffer
;       value   - value to convert
;       padding - type of padding (0=none, 1=spaces, 2=zeroes)
;       padlen  - minimal length of number. ignored if padding=0.
;       plus    - 0=no "+", 1="+" for positive numbers
;                 2=" " for positive, only useful when padding=2
;       len     - address of dword variable, which will on return hold number
;                 of characters of converted number. ignored if 0.
; ret: CF=1 on error
; error: ERR_ZERO_SIZE    - destlen=0. len set to 0
;        ERR_INVALID_MODE - bad padding type. len set to 0
;        ERR_BUFFER_FULL  - destination buffer small. only first destlen
;                           characters of result were stored. len=destlen.
;=============================================================================
proc extconv.id2dec dest, destlen, value, padding, padlen, plus, len
	push	ebx ecx edx edi esi

	cmp	[destlen], 0
	je	.error_destlen_zero
	cmp	[padding], 2
	ja	.error_padding_type
	cmp	[plus], 2
	ja	.error_plus

	mov	edi, [dest]

	;EAX = abs(value)
	mov	eax, [value]		;eax = value
	cmp	eax, 0
	jge	@f
	neg	eax
	@@:

	;ECX = number of digits on stack
	;convert value (push remainders to stack as dwords)
	xor	ecx, ecx		;ecx = number of chars
	mov	ebx, 10
.push_digit:
	xor	edx, edx
	div	ebx
	push	edx
	inc	ecx
	cmp	eax, 0
	jnz	.push_digit

	;EBX = destlen
	mov	ebx, [destlen]

.sign1:

	;get padding type
	cmp	[padding], 0
	je	.padded
	mov	dl, ' '
	cmp	[padding], 1
	je	@f
	mov	dl, '0'
	@@:

	;if padding type is "0", we must handle sign before padding
	cmp	dl, '0'
	jne	.do_padding

	;if value is negative, print '-'
	cmp	[value], 0
	jge	@f
	dec	ebx
	jz	.error_buffer_full
	mov	byte [edi], '-'
	inc	edi
	jmp	.do_padding
	@@:

	;value is positive, plus=0, don't print anything
	cmp	[plus], 0
	je	.do_padding

	;value is positive, print ' ' if plus=2
	cmp	[plus], 2
	jne	@f
	dec	ebx
	jz	.error_buffer_full
	mov	byte [edi], ' '
	inc	edi
	jmp	.do_padding
	@@:

	;value is positive, print '+' if plus=1
	cmp	[plus], 1
	jne	@f
	dec	ebx
	jz	.error_buffer_full
	mov	byte [edi], '+'
	inc	edi
	jmp	.do_padding
	@@:

	;do padding
	;EDI = dest, EBX=destlen, ECX=number of digits, DL=padding char
.do_padding:

	;padding disabled?
	cmp	[padding],0
	je	.padded

	;ESI = number of characters to pad
	mov	esi, ecx

	;if padding type=2, sign doesn't count into padding
	cmp	[padding],2
	je	.remove_sign_from_padding_done

	;if value<0 or plus!=0, increase ESI by one
	cmp	[value], 0
	jl	@f
	cmp	[plus], 0
	ja	@f
	jmp	.remove_sign_from_padding_done
	@@:
	inc	esi
.remove_sign_from_padding_done:

	;pad ESI up to padlen
	mov	eax, [padlen]
.pad:	cmp	eax, esi
	jbe	.padded
	mov	byte [edi], dl
	inc	edi
	dec	ebx
	jz	.error_buffer_full
	dec	eax
	jmp	.pad
.padded:

.sign2:
	;handle sign, if padding type is other than "0"
	cmp	[padding], 2
	je	.sign_done2

	;if value is negative, print '-'
	cmp	[value], 0
	jge	@f
	dec	ebx
	jz	.error_buffer_full
	mov	byte [edi], '-'
	inc	edi
	jmp	.sign_done2
	@@:

	;value is positive, print '+' if plus=1
	cmp	[plus], 1
	jne	.sign_done2
	dec	ebx
	jz	.error_buffer_full
	mov	byte [edi], '+'
	inc	edi

.sign_done2:

	;pop values from stack and store them
	;EDI = dest, EBX=destlen, ECX=number of digits on stack
.store_digit:
	pop	eax
	mov	al, [conv.hextable + eax]
	mov	[edi], al
	inc	edi
	dec	ebx
	jz	.error_buffer_full
	dec	ecx
	jnz	.store_digit

	;store ending zero
	mov	byte [edi], 0

	;save len
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	sub	edi, [dest]
	mov	[eax], edi
	@@:

	clc

.r:	pop	esi edi edx ecx ebx
	ret
.rc:	stc
	jmp	.r

.error_buffer_full:
	;restore stack
	lea	esp, [esp + 4*(ecx-1)]

	;set len to destlen
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	push	[destlen]
	pop	dword [eax]
	@@:

	mov	eax, ERR_BUFFER_FULL
	jmp	.rc

.error_destlen_zero:
	;set len to 0
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], 0
	@@:

	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_plus:
.error_padding_type:
	;set len to 0
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], 0
	@@:

	mov	eax, ERR_INVALID_MODE
	jmp	.rc
endp



;=============================================================================
; extconv.hex2b
; desc: converts hex string (case insensitive) to unsigned byte (8bit) value
; args: string  - pointer to string which we are converting
;       strlen  - size of string buffer
;       len - pointer to dword variable, which will be filled with length
;                 of converted number. ignored if 0
; ret: CF=1 on error, otherwise
;      EAX = value (AL zero extended to EAX)
; note: converts hex digits until reaches non-digit, or value overflows
; error: ERR_ZERO_SIZE        - strlen=0. len set to 0
;        ERR_OUT_OF_RANGE     - value is > 0FFh
;                               len set to number of chars until overflow
;        ERR_NOT_HEXADECIMAL  - first character is not hex digit
;                               len set to 0
;=============================================================================
proc extconv.hex2b string, strlen, len
	libcall extconv.hex2d_size, [string], [strlen], 0F0h, [len]
	ret
endp



;=============================================================================
; extconv.hex2w
; desc: converts hex string (case insensitive) to unsigned word (16bit) value.
; args: string  - pointer to string which we are converting
;       strlen  - size of string buffer
;       len - pointer to dword variable, which will be filled with length
;                 of converted number. ignored if 0.
; ret: CF=1 on error, otherwise
;      EAX = value (AX zero extended to EAX)
; note: converts hex digits until reaches non-digit, or value overflows
; error: ERR_ZERO_SIZE        - strlen=0. len set to 0
;        ERR_OUT_OF_RANGE     - value is > 0FFFFh.
;                               len set to number of chars until overflow
;        ERR_NOT_HEXADECIMAL  - first character is not hex digit.
;                               len set to 0
;=============================================================================
proc extconv.hex2w string, strlen, len
	libcall extconv.hex2d_size, [string], [strlen], 0F000h, [len]
	ret
endp



;=============================================================================
; extconv.hex2d
; desc: converts hex string (case insensitive) to unsigned dword (32bit) value.
; args: string  - pointer to string which we are converting
;       strlen  - size of string buffer
;       len - pointer to dword variable, which will be filled with length
;                 of converted number. ignored if 0.
; ret: CF=1 on error, otherwise
;      EAX = value
; note: converts hex digits until reaches non-digit, or value overflows
; error: ERR_ZERO_SIZE        - strlen=0. len set to 0
;        ERR_OUT_OF_RANGE     - value is > 0FFFFFFFFh.
;                               len set to number of chars until overflow
;        ERR_NOT_HEXADECIMAL  - first character is not hex digit.
;                               len set to 0
;=============================================================================
proc extconv.hex2d string, strlen, len
	libcall extconv.hex2d_size, [string], [strlen], 0F0000000h, [len]
	ret
endp



;=============================================================================
; extconv.dec2b
; desc: converts decimal string to unsigned byte (8bit) value.
; args: string  - pointer to string which we are converting
;       strlen  - size of string buffer
;       len - pointer to dword variable, which will be filled with length
;                 of converted number. ignored if 0.
; ret: CF=1 on error, otherwise
;      EAX = value
; note: converts hex digits until reaches non-digit, or value overflows
; error: ERR_ZERO_SIZE    - strlen=0. len set to 0
;        ERR_OUT_OF_RANGE - value is > 255
;                           len set to number of chars until overflow
;        ERR_NOT_DECIMAL  - first character is not decimal digit.
;                           len set to 0
;=============================================================================
proc extconv.dec2b string, strlen, len
	libcall extconv.dec2d_size, [string], [strlen], 255, [len]
	ret
endp



;=============================================================================
; extconv.dec2w
; desc: converts decimal string to unsigned word (16bit) value.
; args: string  - pointer to string which we are converting
;       strlen  - size of string buffer
;       len - pointer to dword variable, which will be filled with length
;                 of converted number. ignored if 0.
; ret: CF=1 on error, otherwise
;      EAX = value
; note: converts hex digits until reaches non-digit, or value overflows
; error: ERR_ZERO_SIZE    - strlen=0. len set to 0
;        ERR_OUT_OF_RANGE - value is > 65536
;                           len set to number of chars until overflow
;        ERR_NOT_DECIMAL  - first character is not decimal digit.
;                           len set to 0
;=============================================================================
proc extconv.dec2w string, strlen, len
	libcall extconv.dec2d_size, [string], [strlen], 65536, [len]
	ret
endp



;=============================================================================
; extconv.dec2d
; desc: converts decimal string to unsigned dword (32bit) value.
; args: string  - pointer to string which we are converting
;       strlen  - size of string buffer
;       len - pointer to dword variable, which will be filled with length
;                 of converted number. ignored if 0.
; ret: CF=1 on error, otherwise
;      EAX = value
; note: converts hex digits until reaches non-digit, or value overflows
; error: ERR_ZERO_SIZE    - strlen=0. len set to 0
;        ERR_OUT_OF_RANGE - value is > 4294967295
;                           len set to number of chars until overflow
;        ERR_NOT_DECIMAL  - first character is not decimal digit.
;                           len set to 0
;=============================================================================
proc extconv.dec2d string, strlen, len
	libcall extconv.dec2d_size, [string], [strlen], 4294967295, [len]
	ret
endp



;=============================================================================
; extconv.hex2d_size
; desc: INTERNAL! ignore this :]
;       performs job for hex2b, hex2w, hex2d procedures
;=============================================================================
proc extconv.hex2d_size string, strlen, mask, len
	push	esi ecx ebx

	;error is strlen=0
	cmp	[strlen], 0
	je	.error_zero_size

	;ESI = string
	mov	esi, [string]

	;error if first char isn't digit
	cmp	byte [esi], '0'
	jb	.error_not_hex
	cmp	byte [esi], '9'
	jbe	@f
	cmp	byte [esi], 'A'
	jb	.error_not_hex
	cmp	byte [esi], 'F'
	jbe	@f
	cmp	byte [esi], 'a'
	jb	.error_not_hex
	cmp	byte [esi], 'f'
	ja	.error_not_hex
@@:

	;EBX = final value, for now 0
	mov	ebx, 0

	;ECX = count of converted chars
	mov	ecx, 0

	;loop
.digit_loop:
	;AL=char
	mov	al, [esi]
	inc	esi

	;end if it isn't hex digit
	cmp	al, '0'
	jb	.done
	cmp	al, '9'
	jbe	.numeric
	cmp	al, 'A'
	jb	.done
	cmp	al, 'F'
	jbe	.upcase_letter
	cmp	al, 'a'
	jb	.done
	cmp	al, 'f'
	ja	.done

.lowcase_letter:
	sub	al, 'a'-10
	jmp	.add_digit

.upcase_letter:
	sub	al, 'A'-10
	jmp	.add_digit

.numeric:
	sub	al, '0'

.add_digit:
	;check overflow
	test	ebx, [mask]
	jnz	.overflow

	;add digit
	shl	ebx,4
	movzx	eax, al
	add	ebx, eax

	;continue loop
	inc	ecx
	cmp	ecx, [strlen]
	jb	.digit_loop
.done:

	;fill len
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	[eax], ecx
	@@:

	;return EBX
	mov	eax, ebx

	clc
.r:	pop	ebx ecx esi
	ret
.rc:	stc
	jmp	.r


.overflow:
	;set len
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], ecx
	@@:

	;return error
	mov	eax, ERR_OUT_OF_RANGE
	jmp	.rc

.error_zero_size:
	;set len to 0
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], 0
	@@:

	;return error
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc


.error_not_hex:
	;set len to 0
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], 0
	@@:

	;return error
	mov	eax, ERR_NOT_HEXADECIMAL
	jmp	.rc
endp



;=============================================================================
; extconv.dec2d_size
; desc: INTERNAL! ignore this :]
;       performs job for dec2b, dec2w, dec2d procedures
;=============================================================================
proc extconv.dec2d_size string, strlen, size, len
	push	esi edx ecx ebx

	;error is strlen=0
	cmp	[strlen], 0
	je	.error_zero_size

	;ESI = string
	mov	esi, [string]

	;error if first char isn't digit
	cmp	byte [esi], '0'
	jb	.error_not_dec
	cmp	byte [esi], '9'
	ja	.error_not_dec

	;EBX = final value, for now 0
	mov	ebx, 0

	;ECX = count of converted chars
	mov	ecx, 0

	;loop
.digit_loop:
	;AL=char
	mov	al, [esi]
	inc	esi

	;end if it isn't hex digit
	cmp	al, '0'
	jb	.done
	cmp	al, '9'
	ja	.done

	;convert to binary digit
	sub	al, '0'
	movzx	eax, al

	;add digit
	push	eax esi
	xor	edx, edx
	mov	esi, 10
	mov	eax, ebx
	mul	esi
	pop	esi
	mov	ebx, eax
	pop	eax
	add	ebx, eax

	;check for overflow
	jc	.overflow
	cmp	edx, 0
	ja	.overflow
	cmp	ebx, [size]
	ja	.overflow

	;continue loop
	inc	ecx
	cmp	ecx, [strlen]
	jb	.digit_loop
.done:

	;fill len
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	[eax], ecx
	@@:

	;return EBX
	mov	eax, ebx

	clc
.r:	pop	ebx ecx edx esi
	ret
.rc:	stc
	jmp	.r


.overflow:
	;set len
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], ecx
	@@:

	;return error
	mov	eax, ERR_OUT_OF_RANGE
	jmp	.rc

.error_zero_size:
	;set len to 0
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], 0
	@@:

	;return error
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc


.error_not_dec:
	;set len to 0
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], 0
	@@:

	;return error
	mov	eax, ERR_NOT_DECIMAL
	jmp	.rc
endp

;by vid