;FASM STANDARD LIBRARY
;Basic definition macros used by library

;arguments:
; equ   SYSTEM          win32 or linux

;contains:
; equ   FASMLIB_VERSION_STRING   "x.y.z"
; equ   FASMLIB_VERSION_MAJOR    x
; equ   FASMLIB_VERSION_MINOR    y
; equ   FASMLIB_VERSION_REVISION z
; macro idata
; macro udata
; macro IncludeIData
; macro IncludeUData
; macro IncludeAllData
; macro gettimedate
; macro ifndef          expr*
; macro libcall         proc*,[arg]
; macro jmp taken       dest
; macro jmp untaken     dest


;define FASMLIB version
FASMLIB_VERSION_MAJOR	 equ 0
FASMLIB_VERSION_MINOR	 equ 5
FASMLIB_VERSION_REVISION equ 0
match v1 v2 v3, FASMLIB_VERSION_MAJOR FASMLIB_VERSION_MINOR FASMLIB_VERSION_REVISION {
  FASMLIB_VERSION_STRING equ `v1#'.'#`v2#'.'#`v3
}

;check operating system
match =SYSTEM,SYSTEM {include "you have to define operating system: SYSTEM equ win32 or SYSTEM equ linux"}

;define macros for compilation errors
__preprocess_error fix macro*
__pass_error fix __undefined_opcode_blahblahblah
__assembly_error fix rb -1

;==============================================
;ifndef
;desc: checks if expression (usually symbol) is not "defined" previously
;      in code (not in whole file, like with "if ~defined expr")
;args: expr - any expression acceptable as argument to "defined" operator
;note: end block with "end if", just like any "if" command
macro ifndef expr*
{
   local ..HERE
   if defined ..HERE | ~ defined expr
     ..HERE = 1
}



;==============================================
;disphex
;desc: debugging macro to display value of hex byte at assembly time
;args: value - byte value to display
macro disphexb value*
{
   local ..b
   ..b = ((value) shr 4) and 0Fh
   if (..b >= 10)
     ..b = ..b+'A'-10
   else
     ..b = ..b+'0'
   end if
   display ..b
     ..b = (value) and 0Fh
   if (..b >= 10)
     ..b = ..b+'A'-10
   else
     ..b = ..b+'0'
   end if
   display ..b
}

;==============================================
;disphex
;desc: debugging macro to display value of hex dwords at assembly time, prepended by text
;args: text  - text to display
;      dword - dword value to display
macro disphex [text1*, value*]
{
  forward
    display text1
    disphexb (((value) shr 18h) and 0FFh)
    disphexb (((value) shr 10h) and 0FFh)
    disphexb (((value) shr 08h) and 0FFh)
    disphexb ((value) and 0FFh)
  common
    display 13,10
}



;==============================================
;gettimedate
;desc: macro to get current date as constants
;      DAY, MONTH, YEAR
;      DATE = date in (EU) string format for easier definition with "db" or "du"
macro gettimedate {
  local TIME

  macro months [dayscount]
  \{
    \forward
     if DAY > dayscount
      DAY = DAY-dayscount
      MONTH = MONTH+1
    \forward
     end if
  \}
  TIME = %T
  DAY = TIME/(24*3600)
  DAY = DAY - (DAY+365)/(3*365+366)
  YEAR = 1970+DAY/365
  DAY = DAY mod 365 + 1
  MONTH = 1
  if YEAR mod 4 = 0
    FEBDAYS=29
  else
    FEBDAYS=28
  end if
  months 31,FEBDAYS,31,30,31,30,31,31,30,31,30,31

  DATE equ (DAY/10)   +'0', (DAY mod 10)  +'0', '.',\
	   (MONTH/10) +'0', (MONTH mod 10)+'0', '.',\
	   (YEAR/1000)+'0', ((YEAR/100) mod 10)+'0', ((YEAR/10) mod 100)+'0', (YEAR mod 10)+'0'
}

;==============================================
;jcc taken, jcc untaken
;desc: conditional jumps with branch hints
;note: words "taken" and "untaken" must be lowcase
irp jmp, je, jz, ja, jb, jae, jbe, jg, jl, jge, jle, jne, jnz, jna, jnb, jnae, jnbe, jng, jnl, jnge, jnle
{
  macro jmp arg
  \{
    is_plain equ 1
    match =untaken dest, arg
    \\{
      cs jmp dest
      restore is_plain
      is_plain equ 0
    \\}
    match =taken dest, arg
    \\{
      ds jmp dest
      restore is_plain
      is_plain equ 0
    \\}
    match =1, is_plain \{
       jmp arg
    \\}
    restore is_plain
  \}
}

;include parts of this module
include 'fasmlib/data.inc'
include 'fasmlib/proc.inc'
include 'fasmlib/err.inc'

;include system dependant part
match os,SYSTEM {
  include `os#'/fasmlib.inc'
}

;by vid