;FASM STANDARD LIBRARY
;Basic error handling definitions

;macro to define error table
macro errtable [name, string]
{
  common
    local ..count
    ..count = 0
    err.table:
  forward
    local ..string
    if used name | INCLUDE_ALL_ERRORS eq 1
      dd ..string
      name = ..count
      ..count = ..count + 1
    end if
  forward
    idata \{
    if used name | INCLUDE_ALL_ERRORS eq 1
      ..string db string,0
    end if
    \}
  common
    err.count = ..count
}

;predefined error codes
errtable \
ERR_FASMLIB			, "Internal FASMLIB assertion failed. This is bug in FASMLIB. Please report",\
ERR_UNKNOWN			, "Unknown error",\
ERR_UNSUPPOERTED		, "(internal) Feature not supported by FASMLIB",\
ERR_MODULE_NOT_INITIALIZED	, "(internal) Module not initialized",\
\
ERR_OUT_OF_MEM			, "Out of memory",\
ERR_HEAP_CORRUPT		, "(internal) Heap corrupt",\
ERR_MEMORY_LEAK 		, "(internal) Memory leak discovered",\
ERR_INVALID_POINTER		, "(internal) Invalid pointer",\
\
ERR_NOT_HEXADECIMAL		, "Awaiting hexadecimal value",\
ERR_NOT_DECIMAL 		, "Awaiting decimal value",\
ERR_NOT_FLOAT			, "Awaiting floating point value",\
ERR_NOT_BLANK			, "Awaiting blank space",\
ERR_OUT_OF_RANGE		, "Out of range",\
ERR_FORMAT_STRING		, "(internal) Bad format string",\
ERR_ZERO_SIZE			, "(internal) Zero size",\
ERR_ZERO_VALUE			, "(internal) Zero value",\
ERR_BUFFER_FULL 		, "(internal) Insufficent buffer space",\
ERR_INVALID_MODE		, "(internal) Invalid mode",\
\
ERR_INVALID_HANDLE		, "(internal) Invalid handle",\
ERR_HANDLE_NOT_WRITEABLE	, "(internal) Handle not writeable",\
ERR_HANDLE_NOT_READABLE 	, "(internal) Handle not readable",\
ERR_HANDLE_NOT_CLOSEABLE	, "(internal) Handle not closeable",\
ERR_HANDLE_NO_BUFFER		, "(internal) Handle doesn't have buffer associated",\
ERR_HANDLE_CORRUPT		, "(internal) Handle is corrupt",\
ERR_HANDLE_EOF			, "End of file reached",\
\
ERR_FILE_EXISTS 		, "File already exists",\
ERR_FILE_NOT_FOUND		, "File not found",\
ERR_CANT_OPEN_FILE		, "Can't open file"


purge errtable


idata
{
err._invalid_error_code db "(internal) Invalid error code passed to err.text!",0

;---------------------------------------------------------------------
; get text from error code
proc err.text errcode

	mov	eax, [errcode]

	;error codes below err.count are defined in table
	cmp	eax, err.count
	ja	@f
	mov	eax, [err.table + 4*eax]
	jmp	.done
@@:

	;other error codes below 1000h are invalid
	cmp	eax, 1000h
	jae	@f
	mov	eax, err._invalid_error_code
	jmp	.done
	@@:

	;external error codes are pointers to error message
.done:
	clc
	ret
endp

;----------------------------------------------------------------------
; definition of external error messages
struc errcode [string] {
  common
  idata \{
    if used .
      . db string,0
    end if
  \}
}