;FASM STANDARD LIBRARY
;file I/O stream procedures

;proc fstream.open              name
;proc fstream.extopen           buffer, name, rbuf, rbuflen
;proc fstream.create            name, mode
;proc fstream.extcreate         buffer, name, mode

;TODO - fstream.append
;TODO - fstream.openhandle


;FSTRM interface
struc FSTRM
{
  . STREAM
  .handle dd ?
}
virtual at 0
  FSTRM FSTRM
  sizeof.FSTRM = $
  match =1, DISPLAY_STRUC_INFO {
    disphex 'size of FSTREAM object: ', sizeof.FSTRM
  }
end virtual


ifndef FSTREAM_WRITE_BUFFER
  FSTREAM_WRITE_BUFFER equ 200h ;512
end if

ifndef FSTREAM_READ_BUFFER
  FSTREAM_READ_BUFFER equ 200h ;512
end if



proc fstream.open name
	libcall fstream.extopen, 0, [name], 1, FSTREAM_READ_BUFFER, 1, FSTREAM_WRITE_BUFFER
	ret
endp



proc fstream.extopen buffer, name, rbuf, rbuflen
	push	ebx edi

	;EBX = file handle
	libcall file.open, [name], 0
	jc	.rc
	mov	ebx, eax

	;EDI = stream object
	libcall stream.create,\
		[buffer],\				;pass info from caller
		dword 'FILE',\				;stream type
		sizeof.FSTRM - sizeof.STREAM,\	      ;size of additional data
		fstream.read_method,\	;read method
		0,\			;non-blocking read method
		0,\			;peek method unsupported  ;TODO: add, by checking if file is ordinary
		0,\			;non-blocking peek method unsupported
		0,\			;skip method unsupported  ;TODO: add it!
		0,\			;write method disabled
		0,\			;no flush method needed
		fstream.close_method,\	;close method
		[rbuf], [rbuflen],\	;pass read buffer info from caller
		0, 0			;no write buffer needed
	jc	.rc
	mov	edi, eax

	;store additional data
	mov	[edi+FSTRM.handle], ebx

	;return stream handle
	mov	eax, edi

.rnc:	clc
.r:	pop	edi ebx
	ret
.rc:	stc
	jmp	.r
endp



proc fstream.create name, mode
	libcall fstream.extcreate, 0, [name], [mode], 1, FSTREAM_WRITE_BUFFER
	ret
endp



proc fstream.extcreate buffer, name, mode, wbuf, wbuflen
	push	ebx edi

	;error if mode >= 2
	cmp	[mode], 2
	jae	.error_invalid_mode

	;EBX = file handle
	libcall file.create, [name], [mode]
	jc	.rc
	mov	ebx, eax

	;EDI = stream object
	libcall stream.create,\
		[buffer],\				;pass info from caller
		dword 'FILE',\				;stream type
		sizeof.FSTRM - sizeof.STREAM,\	      ;size of additional data
		0,\			;read method disabled
		0,\			;non-blocking read method unsupported
		0,\			;peek method unsupported
		0,\			;non-blocking peek method unsupported
		0,\			;skip method unsupported  ;TODO: add it!
		fstream.write_method,\	;write method
		0,\			;no flush method needed
		fstream.close_method,\	;close method
		0, 0,\			;we don't need read buffer with write-only handle
		[wbuf], [wbuflen]	;pass write buffer info from caller
	jc	.rc
	mov	edi, eax

	;store additional data
	mov	[edi+FSTRM.handle], ebx

	;return stream object/handle
	mov	eax, edi

.rnc:	clc
.r:	pop	edi ebx
	ret
.rc:	stc
	jmp	.r

.error_invalid_mode:
	mov	eax, ERR_INVALID_MODE
	jmp	.rc
endp



;internal, use stream.close
proc fstream.close_method stream
	push	ebx
	mov	ebx, [stream]

	;error if this is not file stream
	cmp	[ebx + STREAM.sign], 'STRM'
	jne	.error_invalid_stream
	cmp	[ebx + STREAM.type], 'FILE'
	jne	.error_invalid_stream

	;call file.close
	libcall file.close, [ebx+FSTRM.handle]
	jc	.rc

.rnc:	clc
.r:	pop    ebx
	ret
.rc:	stc
	jmp	.r

.error_invalid_stream:
	mov	eax, ERR_INVALID_HANDLE
	jmp	.rc
endp



;internal, use stream.write
proc fstream.write_method stream, buffer, buflen
	push	ebx
	mov	ebx, [stream]

	;error if this is not file stream
	cmp	[ebx + STREAM.sign], 'STRM'
	jne	.error_invalid_stream
	cmp	[ebx + STREAM.type], 'FILE'
	jne	.error_invalid_stream

	;call file.write
	libcall file.write, [ebx+FSTRM.handle], [buffer], [buflen]
	jc	.rc

	;no bytes unwritten
	mov	eax, 0

.r:	pop	ebx
	ret
.rc:	stc
	jmp	.r

.error_invalid_stream:
	mov	eax, ERR_INVALID_HANDLE
	jmp	.rc
endp



;internal, use "stream.read" to read file stream
proc fstream.read_method stream, buffer, buflen
	push	ebx
	mov	ebx, [stream]

	;error if this is not file stream
	cmp	[ebx + STREAM.sign], 'STRM'
	jne	.error_invalid_stream
	cmp	[ebx + STREAM.type], 'FILE'
	jne	.error_invalid_stream

	;call file.read
	libcall file.read, [ebx+FSTRM.handle], [buffer], [buflen]
	jc	.rc

	;pass OF and EAX

.rnc:	clc
.r:	pop    ebx
	ret
.rc:	stc
	jmp	.r

.error_invalid_stream:
	mov	eax, ERR_INVALID_HANDLE
	jmp	.rc
endp


restore FSTREAM_READ_BUFFER
restore FSTREAM_WRITE_BUFFER