;FASM STANDARD LIBRARY
;Linux stream I/O routines

;contains:
; proc  stream.writebuf stream, buffer, size


;predefined streams:
idata {
  stream.initialized dd 0

  ;TODO - add to doc: these are NOT closeable
  stream.stdout STREAM\
		'STDO', 	       \;type
		0,		       \;read
		0,		       \;peek
		0,		       \;skip
		stream.write_stdout,   \;write
		0,		       \;readnb
		0,		       \;peeknb
		0,		       \;skipnb
		0,		       \;writenb
		0,		       \;close
		2,		       \;readbuftype    = dynamic
		0,		       \;readbuf        = not yet allocated
		0			;readbufsize    = not yet allocated

  stream.stderr STREAM\
		'STDE', 	       \;type
		0,		       \;read
		0,		       \;peek
		0,		       \;skip
		stream.write_stderr,   \;write
		0,		       \;readnb
		0,		       \;peeknb
		0,		       \;skipnb
		0,		       \;writenb
		0,		       \;close
		2,		       \;readbuftype    = dynamic
		0,		       \;readbuf        = not yet allocated
		0			;readbufsize    = not yet allocated

  stream.stdin	STREAM\
		'STDI', 	       \;type
		stream.read_stdin,     \;read
		0,		       \;peek
		0,		       \;skip
		0,		       \;write
		0,		       \;readnb
		0,		       \;peeknb
		0,		       \;skipnb
		0,		       \;writenb
		0,		       \;close
		2,		       \;readbuftype    = dynamic
		0,		       \;readbuf        = not yet allocated
		0			;readbufsize    = not yet allocated
}



;===============================================================================================
; stream.init
; desc: initializes stream module
; args: none
; ret: CF set on error
; error: ERR_UNKNOWN
;===============================================================================================
proc stream.init
	mov	[stream.initialized], 1
	clc
	ret
endp



;===============================================================================================
; stream.init
; desc: initializes stream module
; args: none
; ret: CF set on error
; error: ERR_MODULE_NOT_INITIALIZED
;        ERR_UNKNOWN
;===============================================================================================
proc stream.uninit
	push	ebx ecx edx

	cmp	[stream.initialized], 0
	je	.error_not_initialized

	;flushnut stdin
	mov	edx, 0	    ;flush input queue
	mov	ecx, 540Bh  ;TCFLSH
	mov	ebx, 1	    ;stdin
	mov	eax, 54     ;sys_ioctl
	int	80h
	cmp	eax, -4096
	ja	.error_ioctl

	;to free read buffer
	libcall stream.close, stream.stdin
	jc	.rc

	mov	[stream.initialized], 1
	clc
.r:	pop	edx ecx ebx
	ret
.rc:	stc
	jmp	.r

.error_not_initialized:
	mov	eax, ERR_MODULE_NOT_INITIALIZED
	stc
	jmp	.r

.error_ioctl:
	mov	eax, ERR_UNKNOWN
	stc
	jmp	.r
endp


;===============================================================================================
; stream.write_stdout
; desc: writes (blocking) to system standard output
; args: buffer - pointer to buffer which holds data
;       size - size of buffer (in bytes)
; ret: CF set on error
; note: this procedure is not to be called directly. it's used by stream operations on stream.stdout
; error: ERR_INVALID_HANDLE - invalid stream handle
;        ERR_ZERO_SIZE      - if size=0
;        ERR_UNKNOWN
;===============================================================================================
proc stream.write_stdout, stream, buffer, size
	push	ebx ecx edx

	;check if size=0
	cmp	[size], 0
	ja	@f
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc
	@@:

	;call sys_write()
	mov	edx, [size]
	mov	ecx, [buffer]
	mov	ebx, 1		;stdout
	mov	eax, 4		;sys_write
	int	80h
	cmp	eax, -4096
	ja	.error_write

	;check if all bytes were written
	cmp	eax,[size]
	je	@f
	mov	eax, ERR_UNKNOWN
	jmp	.rc
	@@:

	;return no unwritten bytes
	mov	eax, 0

.rnc:	clc
.r:	pop	edx ecx ebx
	ret
.rc:	stc
	ret

.error_write:

	;TODO: catch invalid handle

	mov	eax, ERR_UNKNOWN
	jmp	.rc
endp



;===============================================================================================
; stream.write_stderr
; desc: writes (blocking) to system standard error
; args: buffer - pointer to buffer which holds data
;       size - size of buffer (in bytes)
; ret: CF set on error
; note: this procedure is not to be called directly. it's used by stream operations on stream.stderr
; error: ERR_INVALID_HANDLE - invalid stream handle
;        ERR_ZERO_SIZE      - if size=0
;        ERR_UNKNOWN
;===============================================================================================
proc stream.write_stderr, stream, buffer, size
	push	ebx ecx edx

	;check if size=0
	cmp	[size], 0
	ja	@f
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc
	@@:

	;call sys_write()
	mov	edx, [size]
	mov	ecx, [buffer]
	mov	ebx, 2		;stderr
	mov	eax, 4		;sys_write
	int	80h
	cmp	eax, -4096
	ja	.error_write

	;check if all bytes were written
	cmp	eax,[size]
	je	@f
	mov	eax, ERR_UNKNOWN
	jmp	.rc
	@@:

	;no unread bytes
	xor	eax, eax

.rnc:	clc
.r:	pop	edx ecx ebx
	ret
.rc:	stc
	ret

.error_write:

	;TODO: catch invalid handle

	mov	eax, ERR_UNKNOWN
	jmp	.rc
endp



;===============================================================================================
; stream.read_stdin
; desc: reads from system standard input. blocks if there isn't enough data typed already.
;       this is used by "libcall stream.read, stream.stdin, ...". Don't call directly unless you
;       really have to
; args: stream - pointer to stream
;       buffer - pointer to buffer which holds data
;       count  - number of bytes to read
; ret: CF set on error, otherwise
;      OF set on EOF, and
;      EAX = number of bytes read  (if OF=0, then EAX=count)
; error: ERR_INVALID_HANDLE - called on other stream than stream.stdin
;        ERR_ZERO_SIZE      - if size=0
;        ERR_FASMLIB        - bug in FASMLIB
;        ERR_UNKNOWN
;===============================================================================================
proc stream.read_stdin, stream, buffer, count
local readen dd ?
	push	ebx ecx edx edi
	pushf

	;check if this is STDIN
	mov	eax, [stream]
	cmp	[eax + STREAM.type], 'STDI'
	jne	.error_invalid_stream

	;error if count=0
	cmp	[count], 0
	je	.error_count_zero

	;EBX = size
	mov	ebx, [count]

	;EDI = buffer
	mov	edi, [buffer]

.read:
	;call sys_read()
	;function returns after reading one line, which can be less than size we want
	push	ebx
	mov	edx, [count]
	mov	ecx, edi
	mov	ebx, 0		;stdin
	mov	eax, 3		;sys_read
	int	80h
	pop	ebx

	;check sys_read() error
	cmp	eax, -4096
	ja	.error_read

	;done if all bytes were readen
	sub	ebx, eax
	jb	.error_readen_more_than_wanted
	je	.done

	;eof if 0 bytes were readen
	cmp	eax, 0
	je	.done

	;continue reading next line
	add	edi, eax
	jmp	.read

.done:
	;EAX = number of bytes readen
	mov	eax, [count]
	sub	eax, ebx

	;set OF if we didn't read all bytes (EOF)
	cmp	ebx, 0
	ja	.ro

.rno:	and	word [esp], not (1 shl 11) ;clear OF
.rnc:	and	byte [esp], not 1 ;clear CF
.r:	popf
	pop	edi edx ecx ebx
	ret
.rc:	or	byte [esp], 1 ;set CF
	jmp	.r
.ro:	or	word [esp], 1 shl 11 ;set OF
	jmp	.rnc

.error_read:

	;TODO catch

	mov	eax, ERR_UNKNOWN
	jmp	.rc

.error_invalid_stream:
	mov	eax, ERR_INVALID_HANDLE
	jmp	.rc

.error_count_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_readen_more_than_wanted:
	mov	eax, ERR_FASMLIB
	jmp	.rc
endp

;by vid