;FASM STANDARD LIBRARY
;memory stream procedures

;proc memstream.open              name
;proc memstream.extopen           buffer, name, rbuf, rbuflen


;MEMRSTRM interface
struc MEMRSTRM \
      pointer, memsize
{
  . STREAM 'MEMR',\
      memstream.read_method, memstream.peek_method, memstream.skip_method, 0,\	;read, peek, skip, write methods
      memstream.read_method, memstream.peek_method, 0,\ 			;nonblocking methods are same
      0, 0,\		;no flush or close method
      0,0,0,0,0,0	;no read or write buffering
  .pointer dd pointer+0
  .memsize dd memsize+0
  align 4
}
virtual at 0
  MEMRSTRM MEMRSTRM
  sizeof.MEMRSTRM = $
  match =1, DISPLAY_STRUC_INFO {
    disphex 'size of MEMRSTRM object: ', sizeof.MEMRSTRM
  }
end virtual


;MEMWSTRM interface
struc MEMWSTRM \
      pointer, memsize
{
  . STREAM 'MEMW',\
      0, 0, 0, memstream.write_method,\  ;read, peek, skip, write methods
      0, 0, memstream.write_method,\	 ;nonblocking methods are same
      0, 0,\		;no flush or close method
      0,0,0,0,0,0	;no read or write buffering
  .pointer dd pointer+0
  .memsize dd memsize+0
  align 4
}
virtual at 0
  MEMWSTRM MEMWSTRM
  sizeof.MEMWSTRM = $
  match =1, DISPLAY_STRUC_INFO {
    disphex 'size of MEMWSTRM object: ', sizeof.MEMWSTRM
  }
end virtual


proc memstream.openr memptr, memsize
	libcall memstream.extopenr, 0, [memptr], [memsize]
	ret
endp



proc memstream.extopenr objbuf, memptr, memsize
	push	ebx edi

	;EDI = stream object
	libcall stream.create,\
		[objbuf],\			      ;pass info from caller
		dword 'MEMR',\			      ;stream type
		sizeof.MEMRSTRM - sizeof.STREAM,\      ;size of additional data
		memstream.read_method,\   ;read method
		memstream.read_method,\   ;non-blocking read method
		memstream.peek_method,\   ;peek method
		memstream.peek_method,\   ;non-blocking peek method
		memstream.skip_method,\   ;skip method
		0,\			  ;write method disabled
		0,\			  ;no flush method needed
		0,\			  ;no close method needed
		0, 0, 0, 0		  ;no buffering
	jc	.rc
	mov	edi, eax

	;store additional data
	mov	eax, [memptr]
	mov	[edi+MEMRSTRM.pointer], eax
	mov	eax, [memsize]
	mov	[edi+MEMRSTRM.memsize], eax

	;return stream handle
	mov	eax, edi

.rnc:	clc
.r:	pop	edi ebx
	ret
.rc:	stc
	jmp	.r

endp



proc memstream.openw memptr, memsize
	libcall memstream.extopenw, 0, [memptr], [memsize]
	ret
endp



proc memstream.extopenw objbuf, memptr, memsize
	push	ebx edi

	;EDI = stream object
	libcall stream.create,\
		[objbuf],\			      ;pass info from caller
		dword 'MEMW',\			      ;stream type
		sizeof.MEMWSTRM - sizeof.STREAM,\      ;size of additional data
		0,\			  ;read method disabled
		0,\			  ;non-blocking read method disabled
		0,\			  ;peek method disabled
		0,\			  ;non-blocking peek method disabled
		0,\			  ;skip method disabled
		memstream.write_method,\  ;write method
		0,\			  ;no flush method needed
		0,\			  ;no close method needed
		0, 0, 0, 0		  ;no buffering
	jc	.rc
	mov	edi, eax

	;store additional data
	mov	eax, [memptr]
	mov	[edi+MEMRSTRM.pointer], eax
	mov	eax, [memsize]
	mov	[edi+MEMRSTRM.memsize], eax

	;return stream handle
	mov	eax, edi

.rnc:	clc
.r:	pop	edi ebx
	ret
.rc:	stc
	jmp	.r

endp



;internal, use "stream.read" to read file stream
proc memstream.read_method stream, buffer, count
	push	ebx ecx
	pushf

	;EBX = stream
	mov	ebx, [stream]
	cmp	[ebx + STREAM.sign], 'STRM'
	jne	.error_invalid_stream
	cmp	[ebx + STREAM.type], 'MEMR'
	jne	.error_invalid_stream

	;ECX = bytes we can read = min(size, stream.memsize)
	mov	ecx, [count]
	mov	eax, [ebx + MEMRSTRM.memsize]
	cmp	ecx, eax
	cmova	ecx, eax

	test	ecx, ecx
	jz	@f
	libcall mem.copy, [buffer], [ebx + MEMRSTRM.pointer], ecx
	jc	.rc
	sub	[ebx + MEMRSTRM.memsize], ecx
	@@:

	;return OF?
	cmp	ecx, [count]
	jb	.ro

.rno:	btr	dword [esp], 11 ;clear OF
.rnc:	btr	dword [esp], 0	;clear CF
.r:	popf
	pop	ecx ebx
	ret
.rc:	bts	dword [esp], 0	;set CF
	jmp	.r
.ro:	bts	dword [esp], 11 ;set OF
	jmp	.rnc

.error_invalid_stream:
	mov	eax, ERR_INVALID_HANDLE
	jmp	.rc
endp



proc memstream.peek_method, stream, count, readlen
	push	ebx ecx
	pushf

	;EBX = stream
	mov	ebx, [stream]
	cmp	[ebx + STREAM.sign], 'STRM'
	jne	.error_invalid_stream
	cmp	[ebx + STREAM.type], 'MEMR'
	jne	.error_invalid_stream

	;ECX = bytes we can peek = min(size, stream.memsize)
	mov	ecx, [count]
	mov	eax, [ebx + MEMRSTRM.memsize]
	cmp	ecx, eax
	cmova	ecx, eax

	;set readlen
	mov	eax, [readlen]
	cmp	eax, 0
	je	@f
	mov	[eax], ecx
	@@:

	;if user didn't set readlen, and we are returning OF, then return 0 in EAX,
	;so he immediately get GPF if he tries to read from buffer.
	cmp	[readlen], 0
	jne	@f
	cmp	ecx, [count]
	jnb	@f
	mov	eax, 0
	@@:

	;return pointer to data
	mov	eax, [ebx + MEMRSTRM.pointer]

	;return OF?
	cmp	ecx, [count]
	jb	.ro

.rno:	btr	dword [esp], 11 ;clear OF
.rnc:	btr	dword [esp], 0	;clear CF
.r:	popf
	pop	ecx ebx
	ret
.rc:	bts	dword [esp], 0	;set CF
	jmp	.r
.ro:	bts	dword [esp], 11 ;set OF
	jmp	.rnc

.error_invalid_stream:
	mov	eax, ERR_INVALID_HANDLE
	jmp	.rc
endp


proc memstream.skip_method, stream, count
	push	ebx ecx
	pushf

	;EBX = stream
	mov	ebx, [stream]
	cmp	[ebx + STREAM.sign], 'STRM'
	jne	.error_invalid_stream
	cmp	[ebx + STREAM.type], 'MEMR'
	jne	.error_invalid_stream

	;ECX = bytes we can skip = min(size, stream.memsize)
	mov	ecx, [count]
	mov	eax, [ebx + MEMRSTRM.memsize]
	cmp	ecx, eax
	cmova	ecx, eax

	;skip them
	add	[ebx + MEMRSTRM.pointer], ecx
	sub	[ebx + MEMRSTRM.memsize], ecx

	;return OF?
	cmp	ecx, [count]
	jb	.ro

.rno:	btr	dword [esp], 11 ;clear OF
.rnc:	btr	dword [esp], 0	;clear CF
.r:	popf
	pop	ecx ebx
	ret
.rc:	bts	dword [esp], 0	;set CF
	jmp	.r
.ro:	bts	dword [esp], 11 ;set OF
	jmp	.rnc

.error_invalid_stream:
	mov	eax, ERR_INVALID_HANDLE
	jmp	.rc
endp



proc memstream.write_method stream, buffer, count
	push	ebx ecx
	pushf

	;EBX = stream
	mov	ebx, [stream]
	cmp	[ebx + STREAM.sign], 'STRM'
	jne	.error_invalid_stream
	cmp	[ebx + STREAM.type], 'MEMW'
	jne	.error_invalid_stream

	;ECX = bytes we can write = min(size, stream.memsize)
	mov	ecx, [count]
	mov	eax, [ebx + MEMRSTRM.memsize]
	cmp	ecx, eax
	cmova	ecx, eax

	;write
	libcall mem.copy, [ebx + MEMWSTRM.pointer], [buffer], ecx
	jc	.rc
	add	[ebx + MEMWSTRM.pointer], ecx
	sub	[ebx + MEMWSTRM.memsize], ecx

	;return ERR_BUFFER_FULL if we didn't have enough place to write everything
	cmp	ecx, [count]
	jb	.error_buffer_full

	;all written
	mov	eax, 0

.rno:	btr	dword [esp], 11 ;clear OF
.rnc:	btr	dword [esp], 0	;clear CF
.r:	popf
	pop	ecx ebx
	ret
.rc:	bts	dword [esp], 0	;set CF
	jmp	.r
.ro:	bts	dword [esp], 11 ;set OF
	jmp	.rnc

.error_invalid_stream:
	mov	eax, ERR_INVALID_HANDLE
	jmp	.rc

.error_buffer_full:
	mov	eax, ERR_BUFFER_FULL
	jmp	.rc
endp