;FASM STANDARD LIBRARY
;stream routines for working with text

;TODO: add padding char to text.extwrite.

;contains:

; proc text.write               stream, string
; proc text.write.str           stream, string, buflen
; proc text.write.format        stream, formatstr, [...]
; proc text.write.b2hex         stream, value
; proc text.write.w2hex         stream, value
; proc text.write.d2hex         stream, value
; proc text.write.b2dec         stream, value
; proc text.write.w2dec         stream, value
; proc text.write.d2dec         stream, value
; proc text.write.ib2hex        stream, value
; proc text.write.iw2hex        stream, value
; proc text.write.id2hex        stream, value
; proc text.write.ib2dec        stream, value
; proc text.write.iw2dec        stream, value
; proc text.write.id2dec        stream, value

; proc text.read.char           stream
; proc text.read.chars          stream, buffer, buflen, charlist
; proc text.read.until          stream, buffer, buflen, charlist
; proc text.read.line           stream, buffer, buflen
; proc text.read.hex2b          stream
; proc text.read.hex2w          stream
; proc text.read.hex2d          stream          (alias text.read.hex)
; proc text.read.dec2b          stream
; proc text.read.dec2w          stream
; proc text.read.dec2d          stream          (alias text.read.dec)

; proc text.fetch               stream, string
; proc text.fetch.str           stream, string, buflen
; proc text.fetch.char          stream, char

; proc text.skip.chars          stream, chars
; proc text.skip.until          stream, chars
; proc text.skip.dec            stream
; proc text.skip.hex            stream

; proc text.extwrite.b2hex      stream, value, padding, padlen
; proc text.extwrite.w2hex      stream, value, padding, padlen
; proc text.extwrite.d2hex      stream, value, padding, padlen
; proc text.extwrite.b2dec      stream, value, padding, padlen
; proc text.extwrite.w2dec      stream, value, padding, padlen
; proc text.extwrite.d2dec      stream, value, padding, padlen
; proc text.extwrite.ib2hex     stream, value, padding, padlen, plus
; proc text.extwrite.iw2hex     stream, value, padding, padlen, plus
; proc text.extwrite.id2hex     stream, value, padding, padlen, plus
; proc text.extwrite.ib2dec     stream, value, padding, padlen, plus
; proc text.extwrite.iw2dec     stream, value, padding, padlen, plus
; proc text.extwrite.id2dec     stream, value, padding, padlen, plus

; proc text.extread.chars       stream, buffer, buflen, charlist, maxlen, len
; proc text.extread.until       stream, buffer, buflen, charlist, maxlen, len
; proc text.extread.line        stream, buffer, buflen, maxlen, len
; proc text.extread.hex2b       stream, maxlen, len
; proc text.extread.hex2w       stream, maxlen, len
; proc text.extread.hex2d       stream, maxlen, len     (has alias text.extread.hex)
; proc text.extread.dec2b       stream, maxlen, len
; proc text.extread.dec2w       stream, maxlen, len
; proc text.extread.dec2d       stream, maxlen, len     (has alias text.extread.dec)

; proc text.extskip.chars       stream, chars, maxlen
; proc text.extskip.until       stream, chars, maxlen
; proc text.extskip.dec         stream, maxlen
; proc text.extskip.hex         stream, maxlen



idata {

;list of hex characters. continues with dec characters :)
if used text.hex_chars
  text.hex_chars db "ABCDEFabcdef"
end if

;list of dec characters
if used text.hex_chars | used text.dec_chars
  text.dec_chars db "0123456789",0
end if

;list if blank characters
if used text.blank_chars
  text.blank_chars db 9, ' '
end if

;list of end-of-line chars
if used text.eol_chars | used text.blank_chars
  text.eol_chars db 10,13,0
end if
}


;=============================================================================
; text.write
; desc: writes string to stream
; args: stream - handle of stream to write to
;       string - string which to write
; ret: CF set on error
; note: this procedure uses string without knowing it's buffer length. You
;       should use it only to write string constants. For writing string
;       buffers, use text.write.str
; error: +str.len
;        +stream.write
;=============================================================================
proc text.write stream, string
	libcall stream.write, [stream], [string], <libcall str.len, [string], -1>
	ret
endp



;=============================================================================
; text.write.str
; desc: writes string buffer to stream
; args: stream - handle of stream to write to
;       string - string which to write
;       buflen - length of string buffer
; ret: CF set on error
; note: this procedure uses string without knowing it's buffer length. This is
;       to make it easy to use it with string constants. Be sure about passing
;       only properly zero-terminated strings
; error: +str.len
;        +stream.write
;=============================================================================
proc text.write.str stream, string, buflen
	libcall stream.write, [stream], [string], <libcall str.len, [string], [buflen]>
	ret
endp



;=============================================================================
; text.write.format
; desc: writes format string to stream, with special % values substituted
;       by converted values of additional arguments
; args: stream - handle of stream to write to
;       formatstr - format string, see end of this document for an example
;       [...] - other arguments, depending on format string
; ret: CF set on error
; note: You don't have to clean arguments from stack after calling this.
;       Number of arguments is determined from format string, and they are
;       cleaned by procedure.
; error: ERR_FORMAT_STRING - bad format string
;        +extconv.dec2d
;        +str.len
;        +stream.write.char
;        +stream.write.str
;        +text.write
;        +text.extwrite.d2dec
;        +text.extwrite.id2dec
;        +text.extwrite.d2hex
;        +text.extwrite.id2hex
;=============================================================================
text.write.format:
label .arg1 dword at ebp+16
label .formatstr dword at ebp+12
label .stream  dword at ebp+8
label .convlen dword at ebp-4
label .signed  byte at ebp-5
label .pointed byte at ebp-6
label .size    byte at ebp-7
label .padtype dword at ebp-12
label .padlen  dword at ebp-16
	push	ebp
	mov	ebp, esp
	sub	esp, 16
	push	ebx ecx edx esi edi

	;edi = stream
	mov	edi, [.stream]

	;esi = format string
	mov	esi, [.formatstr]

	;edx = first vararg
	lea	edx, [.arg1]

.format_string_loop:

	;get char
	lodsb

	;check terminating zero
	cmp	al, 0
	je	.done

	;check special char
	cmp	al, '%'
	je	.special_char

	;check escape char (needed for outputting "%")
	cmp	al, '\'
	je	.escaped_char

	jmp	.normal_char

.special_char:

	;set default padding length
	mov	[.padlen], 0		;padding length
	mov	[.padtype], 0		;padding char
	mov	[.pointed], 0		;use pointer
	mov	[.signed], 0		;unsigned by default
	mov	[.size], 4		;size=dword

	;get next character
	lodsb

	;check for 0
	cmp	al, 0
	je	.error_format_string_end

.check_0_padding:

	;check for '0' padding prefix
	cmp	al, '0'
	jne	@f
	mov	[.padtype],2
	lodsb
	@@:

.check_padding_size:

	;check if we are on padding size prefix
	cmp	al, '0'
	jb	.get_pointed
	cmp	al, '9'
	ja	.get_pointed
	dec	esi
	lea	eax, [.convlen]
	libcall extconv.dec2d, esi, -1, eax
	jc	.rc
	mov	[.padlen], eax
	add	esi, [.convlen]
	lodsb
	cmp	al, 0
	je	.error_format_string_end

	;if not padding type set yet, set it to default
	cmp	[.padtype], 0
	jne	@f
	mov	[.padtype], 1
	@@:

.get_pointed:
	cmp	al, 'p'
	jne	@f
	mov	[.pointed], 1
	lodsb
	cmp	al, 0
	je	.error_format_string_end
	@@:

.get_signed_override:
	cmp	al, 'i'
	jne	@f
	mov	[.signed], 1
	lodsb
	cmp	al, 0
	je	.error_format_string_end
	@@:

.get_size_override:
	mov	[.size], 1
	cmp	al, 'b'
	je	@f
	mov	[.size], 2
	cmp	al, 'w'
	je	@f
	mov	[.size], 4
	jmp	.get_type
	@@:
	lodsb

.get_type:

	;check if it's decimal
	cmp	al, 'd'
	je	.print_decimal

	;check if it's hexadecimal
	cmp	al, 'h'
	je	.print_hexadecimal
	cmp	al, 'x'
	je	.print_hexadecimal

	;check if it's string constant
	cmp	al, 's'
	je	.print_string_constant

	;check if it's string
	cmp	al, 'S'
	je	.print_string

	; 'c' for char

	;unknown type char
	add	edx, 4	;for '%' we are processing now
	jmp	.error_format_string


.print_string:

	;get arguments
	mov	ecx, dword [edx+4]  ;string buffer length
	mov	ebx, dword [edx]    ;string pointer
	add	edx, 8

	;check prefixes
	cmp	[.size], 4
	jne	.error_format_string
	cmp	[.pointed], 0
	jne	.error_format_string
	cmp	[.padtype], 0
	jne	.error_format_string

	;get length to pad
	cmp	[.padlen], 0
	je	@f
	libcall str.len, ebx, ecx
	jc	.rc
	sub	[.padlen], eax
	jnc	@f
	mov	[.padlen], 0   ;if string is longer than padding
	@@:

	;display padding
@@:	cmp	[.padlen], 0
	je	@f
	dec	[.padlen]
	libcall stream.writechar, edi, dword ' '
	jc	.rc
	jmp	@b
	@@:

	;display string
	libcall text.write.str, edi, ebx, ecx
	jc	.rc

	jmp	.format_string_loop


.print_string_constant:

	;get arguments
	mov	ebx, dword [edx]    ;string pointer
	add	edx, 4

	;check prefixes
	cmp	[.size], 4
	jne	.error_format_string
	cmp	[.pointed], 0
	jne	.error_format_string
	cmp	[.padtype], 0
	jne	.error_format_string

	;get length to pad
	cmp	[.padlen], 0
	je	@f
	libcall str.len, ebx, ecx
	jc	.rc
	sub	[.padlen], eax
	jnc	@f
	mov	[.padlen], 0   ;if string is longer than padding
	@@:

	;display padding
@@:	cmp	[.padlen], 0
	je	@f
	dec	[.padlen]
	libcall stream.writechar, edi, dword ' '
	jc	.rc
	jmp	@b
	@@:

	;display string
	libcall text.write, edi, ebx
	jc	.rc

	jmp	.format_string_loop

.print_hexadecimal:

	;get argument
	mov	ebx, [edx]
	add	edx, 4
	cmp	[.pointed], 0
	je	taken .print_hexadecimal_gotten
	cmp	[.size], 1
	jne	untaken @f
	mov	bl, byte [ebx]
	jmp	.print_hexadecimal_gotten
	@@:
	cmp	[.size], 2
	jne	taken @f
	mov	bx, word [ebx]
	jmp	.print_hexadecimal_gotten
	@@:
	mov	ebx, dword [ebx]
	.print_hexadecimal_gotten:

	;check byte override
	cmp	[.size], 1
	jne	taken @f
	movzx	ebx, bl
	cmp	[.signed], 1
	jne	@f
	movsx	ebx, bl
	@@:

	;check word override
	cmp	[.size], 2
	jne	taken @f
	movzx	ebx, bx
	cmp	[.signed], 2
	jne	@f
	movsx	ebx, bx
	@@:

	;display
	cmp	[.signed], 1
	je	.print_hex_signed
	libcall text.extwrite.d2hex, edi, ebx, [.padtype], [.padlen]
	jc	.rc
	jmp	@f
.print_hex_signed:
	libcall text.extwrite.id2hex, edi, ebx, [.padtype], [.padlen], 2
	jc	.rc
@@:

	jmp	.format_string_loop


.print_decimal:

	;get argument (ugly)
	mov	ebx, [edx]
	add	edx, 4
	cmp	[.pointed], 0
	je	.print_decimal_gotten
	cmp	[.size], 1
	jne	@f
	mov	bl, byte [ebx]
	jmp	.print_decimal_gotten
	@@:
	cmp	[.size], 2
	jne	@f
	mov	bx, word [ebx]
	jmp	.print_decimal_gotten
	@@:
	mov	ebx, dword [ebx]
	@@:
	.print_decimal_gotten:

	;check byte override
	cmp	[.size], 1
	jne	@f
	movzx	ebx, bl
	cmp	[.signed], 1
	jne	@f
	movsx	ebx, bl
	@@:

	;check word override
	cmp	[.size], 2
	jne	@f
	movzx	ebx, bx
	cmp	[.signed], 2
	jne	@f
	movsx	ebx, bx
	@@:

	;display
	cmp	[.signed], 1
	je	.print_dec_signed
	libcall text.extwrite.d2dec, edi, ebx, [.padtype], [.padlen]
	jc	.rc
	jmp	@f
.print_dec_signed:
	libcall text.extwrite.id2dec, edi, ebx, [.padtype], [.padlen], 2
	jc	.rc
@@:

	jmp	.format_string_loop


.escaped_char:

	;escaped char - load next char
	lodsb

	;error if it is end of string
	cmp	al, 0
	je	.error_format_string_end

.normal_char:

	;plain character, just write it
	libcall stream.writechar, edi, eax
	jc	.rc

.continue_format_string_loop:
	jmp	.format_string_loop

.done:

.rnc:	clc
.r:	lea	eax, [edx-4]		;eax = pointer to last argument
	pop	edi esi edx ecx ebx	;restore regs
	mov	esp, ebp      ;remove local variables
	pop	ebp	      ;restore ebp
	pop	dword [eax]   ;copy return address to place of last argument
	mov	esp, eax      ;move stack pointer there
	retn		      ;and return

.rpop:	pop	dword [edx-8]		;re-save value of eax from stack to 2nd last argument
	lea	eax, [edx-8]		;eax = pointer to 2nd last argument
	pop	edi esi edx ecx ebx	;restore regs
	mov	esp, ebp      ;remove local variables
	pop	ebp	      ;restore ebp
	pop	dword [eax+4] ;copy return adress to place of last argument
	mov	esp, eax      ;move stack pointer to saved EAX
	pop	eax	      ;restore EAX
	retn		      ;and return

	jmp	.r

.rc:
	;we must count non-escaped % chars in format string to skip remaining args
	push	eax  ;save error code

.onerror_skip_nextchar:
	lodsb

	;char 0 = end of string = we can finally exit
	cmp	al, 1
	jb	.rpop	;jb=jc

	;handle %
	cmp	al, '%'
	je	.onerror_percent

	;handle escaped character
	cmp	al, '\'
	jne	.onerror_skip_nextchar
	lodsb
	cmp	al, 0
	je	.error_format_string_onerror
	jmp	.onerror_skip_nextchar

.onerror_percent:
	lodsb

	;error if end-of-string
	cmp	al,0
	je	.error_format_string_onerror

	;skip all numbers after "%"
	cmp	al, '0'
	jb	@f
	cmp	al, '9'
	jbe	.onerror_percent

	;skip 'p'
	cmp	al, 'p'
	jne	@f
	lodsb
	cmp	al, 0
	je	.error_format_string_onerror
	@@:

	;skip 'i'
	cmp	al, 'i'
	jne	@f
	lodsb
	cmp	al, 0
	je	.error_format_string_onerror
	@@:

	;skip 'b'
	cmp	al, 'b'
	jne	@f
	lodsb
	cmp	al, 0
	je	.error_format_string_onerror
	@@:

	;skip 'w'
	cmp	al, 'w'
	jne	@f
	lodsb
	cmp	al, 0
	je	.error_format_string_onerror
	@@:

	;decide based on type
	cmp	al, 'S'
	jne	@f		;treat any other type as dword sized  ;TODO - change error on stack
	add	edx, 4
	@@:
	add	edx, 4
	jmp	.onerror_skip_nextchar


	;error in format string
.error_format_string:
	mov	eax, ERR_FORMAT_STRING
	jmp	.rc

	;error after reaching end of string
.error_format_string_end:
	push	ERR_FORMAT_STRING
	stc
	jmp	.rpop

	;error during parsing format string after error :)
	;only when end of string is reached
.error_format_string_onerror:
	mov	dword [esp], ERR_FORMAT_STRING
	stc
	jmp	.rpop



;=============================================================================
; text.write.b2hex
; desc: writes unsigned byte (8bit) to stream as hex string
; args: stream - handle of stream to write to
;       value  - low order byte of this is value to write
; ret: CF set on error
; error: +extconv.b2hex
;        +stream.write
;=============================================================================
proc text.write.b2hex stream, value
local	len	dd ?
local	buf	rb 3

	libcall extconv.b2hex, addr buf, 3, [value], 0, 0, addr len
	jc	.rc

	libcall stream.write, [stream], addr buf, [len]
	jc	.rc

.r:	ret
.rc:	stc
	jmp	.r
endp



;=============================================================================
; text.write.w2hex
; desc: writes unsigned word (16bit) to stream as hex string
; args: stream - handle of stream to write to
;       value  - low order word of this is value to write
; ret: CF set on error
; error: +extconv.w2hex
;        +stream.write
;=============================================================================
proc text.write.w2hex stream, value
local	len	dd ?
local	buf	rb 5

	libcall extconv.w2hex, addr buf, 5, [value], 0, 0, addr len
	jc	.rc

	libcall stream.write, [stream], addr buf, [len]
	jc	.rc

.r:	ret
.rc:	stc
	jmp	.r
endp



;=============================================================================
; text.write.d2hex
; desc: writes unsigned dword (32bit) to stream as hex string
; args: stream - handle of stream to write to
;       value  - value to write
; ret: CF set on error
; error: +extconv.d2hex
;        +stream.write
;=============================================================================
proc text.write.d2hex stream, value
local	len	dd ?
local	buf	rb 9

	libcall extconv.d2hex, addr buf, 9, [value], 0, 0, addr len
	jc	.rc

	libcall stream.write, [stream], addr buf, [len]
	jc	.rc

.r:	ret
.rc:	stc
	jmp	.r
endp



;=============================================================================
; text.write.b2dec
; desc: writes unsigned byte (8 bit) to stream as decimal string
; args: stream - handle of stream to write to
;       value  - low order word of this is value to write
; ret: CF set on error
; error: +extconv.b2dec
;        +stream.write
;=============================================================================
proc text.write.b2dec stream, value
local	len	dd ?
local	buf	rb 3

	libcall extconv.b2dec, addr buf, 3, [value], 0, 0, addr len
	jc	.rc

	libcall stream.write, [stream], addr buf, [len]
	jc	.rc

.r:	ret
.rc:	stc
	jmp	.r
endp



;=============================================================================
; text.write.w2dec
; desc: writes unsigned word (16 bit) to stream as decimal string
; args: stream - handle of stream to write to
;       value  - low order word of this is value to write
; ret: CF set on error
; error: +extconv.w2dec
;        +stream.write
;=============================================================================
proc text.write.w2dec stream, value
local	len	dd ?
local	buf	rb 6

	libcall extconv.w2dec, addr buf, 6, [value], 0, 0, addr len
	jc	.rc

	libcall stream.write, [stream], addr buf, [len]
	jc	.rc

.r:	ret
.rc:	stc
	jmp	.r
endp



;=============================================================================
; text.write.d2dec
; desc: writes unsigned dword (32 bit) to stream as decimal string
; args: stream - handle of stream to write to
;       value  - value to write
; ret: CF set on error
; error: +extconv.d2dec
;        +stream.write
;=============================================================================
proc text.write.d2dec stream, value
local	len	dd ?
local	buf	rb 11

	libcall extconv.d2dec, addr buf, 11, [value], 0, 0, addr len
	jc	.rc

	libcall stream.write, [stream], addr buf, [len]
	jc	.rc

.r:	ret
.rc:	stc
	jmp	.r
endp



;=============================================================================
; text.write.ib2hex
; desc: writes signed byte (8bit) to stream as hex string
; args: stream - handle of stream to write to
;       value  - low order byte of this is value to write
; ret: CF set on error
; error: +extconv.ib2hex
;        +stream.write
;=============================================================================
proc text.write.ib2hex stream, value
local	len	dd ?
local	buf	rb 4

	libcall extconv.ib2hex, addr buf, 4, [value], 0, 0, 0, addr len
	jc	.rc

	libcall stream.write, [stream], addr buf, [len]
	jc	.rc

.r:	ret
.rc:	stc
	jmp	.r
endp



;=============================================================================
; text.write.iw2hex
; desc: writes signed word (16bit) to stream as hex string
; args: stream - handle of stream to write to
;       value  - low order word of this is value to write
; ret: CF set on error
; error: +extconv.iw2hex
;        +stream.write
;=============================================================================
proc text.write.iw2hex stream, value
local	len	dd ?
local	buf	rb 6

	libcall extconv.iw2hex, addr buf, 6, [value], 0, 0, 0, addr len
	jc	.rc

	libcall stream.write, [stream], addr buf, [len]
	jc	.rc

.r:	ret
.rc:	stc
	jmp	.r
endp



;=============================================================================
; text.write.id2hex
; desc: writes signed dword (32bit) to stream as hex string
; args: stream - handle of stream to write to
;       value  - value to write
; ret: CF set on error
; error: +extconv.id2hex
;        +stream.write
;=============================================================================
proc text.write.id2hex stream, value
local	len	dd ?
local	buf	rb 10

	libcall extconv.id2hex, addr buf, 10, [value], 0, 0, 0, addr len
	jc	.rc

	libcall stream.write, [stream], addr buf, [len]
	jc	.rc

.r:	ret
.rc:	stc
	jmp	.r
endp



;=============================================================================
; text.write.ib2dec
; desc: writes signed byte (8bit) to stream as hex string
; args: stream - handle of stream to write to
;       value  - low order byte of this is value to write
; ret: CF set on error
; error: +extconv.ib2dec
;        +stream.write
;=============================================================================
proc text.write.ib2dec stream, value
local	len	dd ?
local	buf	rb 5

	libcall extconv.ib2dec, addr buf, 5, [value], 0, 0, 0, addr len
	jc	.rc

	libcall stream.write, [stream], addr buf, [len]
	jc	.rc

.r:	ret
.rc:	stc
	jmp	.r
endp



;=============================================================================
; text.write.iw2dec
; desc: writes signed word (16bit) to stream as decimal string
; args: stream - handle of stream to write to
;       value  - low order word of this is value to write
; ret: CF set on error
; error: +extconv.iw2hex
;        +stream.write
;=============================================================================
proc text.write.iw2dec stream, value
local	len	dd ?
local	buf	rb 7

	libcall extconv.iw2dec, addr buf, 7, [value], 0, 0, 0, addr len
	jc	.rc

	libcall stream.write, [stream], addr buf, [len]
	jc	.rc

.r:	ret
.rc:	stc
	jmp	.r
endp



;=============================================================================
; text.write.id2dec
; desc: writes signed dword (32bit) to stream as decimal string
; args: stream - handle of stream to write to
;       value  - value to write
; ret: CF set on error
; error: +extconv.id2dec
;        +stream.write
;=============================================================================
proc text.write.id2dec stream, value
local	len	dd ?
local	buf	rb 12

	libcall extconv.id2dec, addr buf, 12, [value], 0, 0, 0, addr len
	jc	.rc

	libcall stream.write, [stream], addr buf, [len]
	jc	.rc

.r:	ret
.rc:	stc
	jmp	.r
endp



;=============================================================================
; desc: reads one character from stream
; args: stream - handle of stream to read from
; ret: CF set on error, otherwise
;      OF set if we are on EOF, otherwise
;      EAX = readen character, zero extended to dword
; error: +stream.readchar
;=============================================================================
proc text.read.char stream
	libcall stream.readchar, [stream]
	ret
endp



;=============================================================================
; text.read.chars
; desc: reads from the stream all chars that are in given list of characters.
; args: stream   - stream to skip characters from
;       buffer   - pointer to buffer, where the characters will be stored
;       buflen   - length of buffer
;       charlist - null-terminated string containing characters to skip
; ret: CF=1 on error, otherwise
;      EAX = number of characters skipped, and
;      ZF=1 if EAX=0 (no characters readen)
; note: - character 0 always stops skipping
; error: ERR_ZERO_SIZE   - buflen=0
;        ERR_BUFFER_FULL - buffer is full. only buflen characters readen.
;        ERR_HANDLE_EOF  - EOF reached without reading anything
;        ERR_FASMLIB     - bug in FASMLIB! Please report
;        +text.extread.chars
;=============================================================================
proc text.read.chars stream, buffer, buflen, charlist
	libcall text.extread.chars, [stream], [buffer], [buflen], [charlist], 0, 0
	ret
endp



;=============================================================================
; text.read.until
; desc: reads from stream all chars until character from charlist is reached
; args: stream   - stream to read characters from
;       buffer   - pointer to buffer, where the characters will be stored
;       buflen   - length of buffer
;       charlist - null-terminated string containing characters on which to
;                  stop reading
; ret: CF=1 on error, otherwise
;      EAX = number of characters readen, and
;      ZF=1 if EAX=0 (no characters readen)
; note: - character 0 always stops reading
; error: ERR_ZERO_SIZE   - buflen=0. len is set to 0
;        ERR_BUFFER_FULL - buffer is full. only buflen characters readen.
;                          len is set to buflen
;        ERR_HANDLE_EOF  - EOF reached without reading anything
;        ERR_FASMLIB     - bug in FASMLIB! Please report
;        +text.extread.until
;=============================================================================
proc text.read.until stream, buffer, buflen, charlist
	libcall text.extread.until, [stream], [buffer], [buflen], [charlist], 0, 0
	ret
endp



;=============================================================================
; text.read.line
; desc: reads line from stream. Skips EOL to next line
; args: stream   - stream to read from
;       buffer   - pointer to buffer, where the characters will be stored
;       buflen   - length of buffer
; ret: CF=1 on error, otherwise
;      EAX = number of characters readen
; note: - this function can skip CRLF, not just LF. This way you can use it to
;         read stream.stdin directly, without being overloaded by text stream.
;=============================================================================
proc text.read.line stream, buffer, buflen
	libcall text.extread.line, [stream], [buffer], [buflen], 0, 0
	ret
endp



;=============================================================================
; text.read.hex2b
; desc: reads hexadecimal number to byte (8 bit). it reads hex digits
;       (0-9, a-z, A-Z) and converts them to number, until it reaches non-hex
;       character, or until number overflows.
; args: stream     - handle of stream from which to read
; ret: CF set on error, otherwise
;      EAX = value (byte zero extended to dword)
; note: - before reading, all blank characters are skipped
; error: ERR_HANDLE_EOF         - file pointer is on EOF. len contains 0
;        ERR_OUT_OF_RANGE       - value > 0FFh
;                                 in this case nothing is read from stream.
;        ERR_NOT_HEXADECIMAL    - file pointer not on hex number (first char isn't 0-9, a-z, A-Z)
;                                 in this case, nothing is read from stream. len contains 0
;        ERR_FASMLIB            - bug in FASMLIB! please report it
;        +text.skip.chars
;=============================================================================
proc text.read.hex2b stream
	push	0		;skip blank chars
	push	1		;no len
	push	0		;no maxlen
	push	0x000000F0	;overflow
	push	[stream]	;stream we read from
	call	text.extread.hex2d_size
	ret
endp



;=============================================================================
; text.read.hex2w
; desc: reads hexadecimal number to word (16 bit). reads hex digits
;       (0-9, a-z, A-Z) and converts them to number, until it reaches non-hex
;       character, or until number overflows.
; args: stream     - handle of stream from which to read
; ret: CF set on error, otherwise
;      EAX = value (word zero extended to dword)
; note: - before reading, all blank characters are skipped
; error: ERR_HANDLE_EOF         - file pointer is on EOF. len contains 0
;        ERR_OUT_OF_RANGE       - value > 0FFFFh
;                                 in this case nothing is read from stream.
;        ERR_NOT_HEXADECIMAL    - file pointer not on hex number (first char isn't 0-9, a-z, A-Z)
;                                 in this case, nothing is read from stream.
;        ERR_FASMLIB            - bug in FASMLIB! please report it
;        +text.skip.chars
;=============================================================================
proc text.read.hex2w stream
	libcall text.extread.hex2d_size, [stream], 0x0000F000, 0, 1, 0
	ret
endp



;=============================================================================
; text.read.hex2d
; desc: reads hexadecimal number to dword (32 bit). reads hex digits
;       (0-9, a-z, A-Z) and converts them to number, until it reaches non-hex
;       character, or until number overflows.
; args: stream     - handle of stream from which to read
; ret: CF set on error, otherwise
;      EAX = value
; note: - before reading, all blank characters are skipped
; error: ERR_HANDLE_EOF      - file pointer is on EOF. len contains 0
;        ERR_OUT_OF_RANGE    - value is > 0FFFFFFFFh
;                              in this case nothing is read from stream.
;        ERR_NOT_HEXADECIMAL - first char isn't 0-9, a-z, A-Z
;                              in this case, nothing is read from stream.
;        ERR_FASMLIB         - bug in FASMLIB! please report it
;        +text.skip.chars
;=============================================================================
proc text.read.hex2d stream
	libcall text.extread.hex2d_size, [stream], 0xF0000000, 0, 1, 0
	ret
endp
if used text.read.hex
  label text.read.hex at text.read.hex2d
end if



;=============================================================================
; text.read.dec2b
; desc: reads adecimal number to byte (8 bit). it reads decimal digits (0-9)
;       and converts them to number, until it reaches non-hex character, or
;       until number overflows.
; args: stream     - handle of stream from which to read
; ret: CF set on error, otherwise
;      EAX = value (byte zero extended to dword)
; note: - before reading, all blank characters are skipped
; error: ERR_HANDLE_EOF   - file pointer is on EOF. len contains 0
;        ERR_OUT_OF_RANGE - value is > 255
;                           in this case nothing is read from stream.
;        ERR_NOT_DECIMAL  - first char isn't 0-9
;                           in this case, nothing is read from stream.
;        ERR_FASMLIB      - bug in FASMLIB! please report it
;        +text.skip.chars
;=============================================================================
proc text.read.dec2b stream
	libcall text.extread.dec2d_size, [stream], 255, 0, 1, 0
	ret
endp



;=============================================================================
; text.read.dec2w
; desc: reads adecimal number to word (16 bit). it reads decimal digits (0-9)
;       and converts them to number, until it reaches non-hex character, or
;       until number overflows.
; args: stream     - handle of stream from which to read
; ret: CF set on error, otherwise
;      EAX = value (byte zero extended to dword)
; note: - before reading, all blank characters are skipped
; error: ERR_HANDLE_EOF   - file pointer is on EOF. len contains 0
;        ERR_OUT_OF_RANGE - value is > 65535
;                           in this case nothing is read from stream.
;        ERR_NOT_DECIMAL  - first char isn't 0-9
;                           in this case, nothing is read from stream.
;        ERR_FASMLIB      - bug in FASMLIB! please report it
;        +text.skip.chars
;=============================================================================
proc text.read.dec2w stream
	libcall text.extread.dec2d_size, [stream], 65535, 0, 1, 0
	ret
endp



;=============================================================================
; text.read.dec2d
; desc: reads adecimal number to dword (32 bit). it reads decimal digits (0-9)
;       and converts them to number, until it reaches non-hex character, or
;       until number overflows.
; args: stream     - handle of stream from which to read
; ret: CF set on error, otherwise
;      EAX = value (byte zero extended to dword)
; note: - before reading, all blank characters are skipped
; error: ERR_HANDLE_EOF   - file pointer is on EOF. len contains 0
;        ERR_OUT_OF_RANGE - value is > 4294967295
;                           in this case nothing is read from stream.
;        ERR_NOT_DECIMAL  - first char isn't 0-9
;                           in this case, nothing is read from stream.
;        ERR_FASMLIB      - bug in FASMLIB! please report it
;        +text.skip.chars
;=============================================================================
proc text.read.dec2d stream
	libcall text.extread.dec2d_size, [stream], 4294967295, 0, 1, 0
	ret
endp
if used text.read.dec
  label text.read.dec at text.read.dec2d
end if



;=============================================================================
; text.fetch
; desc: skips specified constant string off the stream. If there is another
;       string on stream, then nothing is read and ZF=1 is returned
; args: stream - handle of stream to read from
;       string - constant string to fetch
; ret: CF set on error, otherwise
;      ZF set if the string was fetched
; note: - this function should be used only on constant strings. For non-const
;         strings use "text.fetch.str".
; error: +text.fetch.str
;=============================================================================
proc text.fetch stream, string
	libcall text.fetch.str, [stream], [string], -1
	ret
endp



;=============================================================================
; text.fetch.str
; desc: skips specified string off the stream. If there is another string
;       on stream, then nothing is read and ZF=1 is returned
; args: stream - handle of stream to read from
;       string - string to fetch
;       buflen - length of string buffer
; ret: CF set on error, otherwise
;      ZF set if the string was fetched
; error: ERR_FASMLIB - Bug in FASMLIB! Please report.
;        +str.len
;        +stream.peeknb
;        +stream.peek
;        +stream.skip
;=============================================================================
proc text.fetch.str stream, string, buflen
local len dd ?

	push	ebx ecx edx esi edi

	;ESI = string
	mov	esi, [string]

	;find length of string (sorry, lazy and slow way)
	libcall str.len, esi, [buflen]
	jc	.rc
	mov	ecx, eax

	;EBX = index of character we are comparing
	xor	ebx, ebx

	;non-blocking peek as much as we need
	libcall stream.peeknb, [stream], ecx, addr len
	jc	.rc
	jo	.not_match
	mov	edx, [len]
	cmp	edx, 0
	ja	.compare

.peek_more:

	;blocking-peek one more character
	lea	eax, [ebx+1]
	libcall stream.peek, [stream], eax, addr len
	jc	.rc
	jo	.not_match
	mov	edx, [len]

	;non-blocking peek as much as we need
	libcall stream.peeknb, [stream], ecx, addr len
	jc	.rc
	jo	.not_match
	mov	edx, [len]

.compare:
	mov	edi, eax

.compare_char:
	;compare character
	mov	al, [edi+ebx]
	cmp	al, [esi+ebx]
	jne	.rnc   ;return ZF=0

	;are we finished
	inc	ebx
	cmp	ebx, ecx
	ja	.error_internal
	je	.fetched

	;continue comparing loop
	cmp	ebx, edx
	ja	.error_internal
	jb	.compare_char

	;we need to peek some more characters
	jmp	.peek_more

.fetched:

	;skip it from stream
	libcall stream.skip, [stream], ecx
	jc	.rc
	jo	.error_internal

	;return ZF
	cmp	eax, eax
	jmp	.rnc

.not_match:

	;set ZF
	cmp	ecx, -1  ;string probably won't be this big :)

.rnc:	clc
.r:	pop	edi esi edx ecx ebx
	ret
.rc:	or	byte [esp], 1 ;set CF
	jmp	.r

.error_internal:
	mov	eax, ERR_FASMLIB
	jmp	.rc

endp



;=============================================================================
; text.fetch.char
; desc: skips specified charactet off the stream. If there is another character
;       on stream, then nothing is read and ZF=1 is returned
; args: stream - handle of stream to read from
;       char   - low order byte of this dword is character to fetch
; ret: CF set on error, otherwise
;      ZF set if the character was fetched
; error: +stream.peekchar
;        +stream.skip
;=============================================================================
proc text.fetch.char stream, char

	;peek character
	libcall stream.peekchar, [stream]
	jc	.r
	jo	.r

	;return ZF=0, if it isn't the one we want
	cmp	al, byte [char]
	jne	.rnc

	;skip the character
	libcall stream.skip, [stream], 1
	jc	.r
	jo	.error_internal

	;return ZF=1
	cmp	al, al

.rnc:	clc
.r:	ret
.rc:	stc
	jmp	.r

.error_internal:
	mov	eax, ERR_FASMLIB
	jmp	.rc
endp



;=============================================================================
; text.skip.chars
; desc: reads from the stream all chars that are in given list of characters.
; args: stream   - stream to skip characters from
;       charlist - null-terminated string containing characters to skip.
; ret: CF=1 on error, otherwise
;      OF=1 if EOF was reached, and
;      EAX = number of characters skipped
; note: - character 0 always stops skipping
; error: ERR_FASMLIB - bug in FASMLIB! Please report
;        +stream.peek
;        +stream.peeknb
;        +stream.skip
;        +str.poschar
;=============================================================================
proc text.skip.chars stream, charlist
	libcall text.extskip.chars, [stream], [charlist], -1
	ret
endp



;=============================================================================
; text.skip.until
; desc: skips from the stream all chars until some character from charlist is
;       reached
; args: stream   - stream to skip characters from
;       charlist - null-terminated string containing characters not to skip
; ret: CF=1 on error, otherwise
;      EAX = number of characters skipped, and
;      ZF=1 if EAX=0 (no characters skipped)
; note: - character 0 always stops skipping
; error: ERR_FASMLIB - bug in FASMLIB! Please report
;        +stream.peek
;        +stream.peeknb
;        +stream.skip
;        +str.poschar
;=============================================================================
proc text.skip.until stream, charlist
	libcall text.extskip.until, [stream], [charlist], -1
	ret
endp



;=============================================================================
; text.skip.hex
; desc: reads hexadecimal number (character 0-9, a-z, A-Z) off the stream
; args: stream   - stream to skip number from
; ret: CF=1 on error, otherwise
;      EAX = number of characters skipped, and
;      ZF=1 if EAX=0 (no characters skipped)
; note: - this skips without error long numbers, that would cause
;         ERR_OUT_OF_RANGE with text.read or text.fetch
; error: +text.extskip.chars
;=============================================================================
proc text.skip.hex stream
	libcall text.extskip.chars, stream.stdin, text.hex_chars, -1
	ret
endp



;=============================================================================
; text.extskip.dec
; desc: skips decimal number (character 0-9) off the stream
; args: stream   - stream to skip number from
; ret: CF=1 on error, otherwise
;      EAX = number of characters skipped, and
;      ZF=1 if EAX=0 (no characters skipped)
; note: - this skips without error long numbers, that would cause
;         ERR_OUT_OF_RANGE with text.read or text.fetch
; error: +text.extskip.chars
;=============================================================================
proc text.skip.dec stream
	libcall text.extskip.chars, stream.stdin, text.dec_chars, -1
	ret
endp



;=============================================================================
; text.extwrite.b2hex
; desc: writes unsigned byte (8bit) to stream as zero-padded hex string
; args: stream - handle of stream to write to
;       value  - low order byte of this is value to write
;       padding - type of padding (0=none, 1=spaces, 2=zeroes)
;       padlen  - minimal length of number. ignored if padding=0.
; ret: CF set on error
; error: ERR_OUT_OF_RANGE - padlen > 4096
;        ERR_FASMLIB      - internal error in FASMLIB, please report
;        +extconv.d2hex
;        +stream.write
;=============================================================================
proc text.extwrite.b2hex stream, value, padding, padlen
	movzx	eax, byte [value]
	libcall text.extwrite.d2hex, [stream], eax, [padding], [padlen]
	ret
endp



;=============================================================================
; text.extwrite.w2hex
; desc: writes unsigned word (16bit) to stream as zero-padded hex string
; args: stream - handle of stream to write to
;       value  - low order word of this is value to write
;       padding - type of padding (0=none, 1=spaces, 2=zeroes)
;       padlen  - minimal length of number. ignored if padding=0.
; ret: CF set on error
; error: ERR_OUT_OF_RANGE - padlen > 4096
;        ERR_FASMLIB      - internal error in FASMLIB, please report
;        +extconv.d2hex
;        +stream.write
;=============================================================================
proc text.extwrite.w2hex stream, value, padding, padlen
	movzx	eax, word [value]
	libcall text.extwrite.d2hex, [stream], eax, [padding], [padlen]
	ret
endp



;=============================================================================
; text.extwrite.d2hex
; desc: writes unsigned dword (32bit) to stream as zero-padded hex string
; args: stream - handle of stream to write to
;       value  - value to write
;       padding - type of padding (0=none, 1=spaces, 2=zeroes)
;       padlen  - minimal length of number. ignored if padding=0.
; ret: CF set on error
; error: ERR_OUT_OF_RANGE - padlen > 4096
;        ERR_FASMLIB      - internal error in FASMLIB, please report
;        +extconv.d2hex
;        +stream.write
;=============================================================================
proc text.extwrite.d2hex stream, value, padding, padlen
local convlen dd ?

	push	ebx ecx

	;ECX = size of needed buffer, at least 16
	mov	ecx, [padlen]
	cmp	[padding], 0
	jne	@f
	xor	ecx, ecx
	@@:
	inc	ecx
	add	ecx, 3
	and	ecx, not 3  ;align to dword
	mov	eax, 16
	cmp	ecx, eax
	cmovb	ecx, eax

	;sanity-check for padlen > 4096, to prevent big allocation on stack
	cmp	ecx, 4096
	ja	.error_padlen_too_big

	;EBX = allocte padlen+1 buffer on stack
	sub	esp, ecx
	mov	ebx, esp

	;convert
	libcall extconv.d2hex, ebx, ecx, [value], [padding], [padlen], addr convlen
	jc	.error_convert

	;display
	libcall stream.write, [stream], ebx, [convlen]
	jc	.error_write

	;remove buffer from stack
	add	esp, ecx

.r:	pop	ebx
	ret
.rc:	stc
	jmp	.r

.error_write:
	add	esp, ecx
	jmp	.rc

.error_padlen_too_big:
	mov	eax, ERR_OUT_OF_RANGE
	jmp	.rc

.error_convert:
	add	esp, ecx
	;if returned error is ERR_BUFFER_FULL, it's FASMLIB bug
	cmp	eax, ERR_BUFFER_FULL
	jne	.rc
	mov	eax, ERR_FASMLIB
	jmp	.rc

endp



;=============================================================================
; text.extwrite.b2dec
; desc: writes unsigned byte (8 bit) to stream as decimal string
; args: stream - handle of stream to write to
;       value  - low order byte of this is value to write
;       padding - type of padding (0=none, 1=spaces, 2=zeroes)
;       padlen  - minimal length of number. ignored if padding=0.
; ret: CF set on error
; error: ERR_OUT_OF_RANGE - padlen > 4096
;        ERR_FASMLIB      - internal error in FASMLIB, please report
;        +extconv.d2dec
;        +stream.write
;=============================================================================
proc text.extwrite.b2dec stream, value, padding, padlen
	movzx	eax, byte [value]
	libcall text.extwrite.d2dec, [stream], eax, [padding], [padlen]
	ret
endp



;=============================================================================
; text.extwrite.w2dec
; desc: writes unsigned word (16 bit) to stream as decimal string
; args: stream - handle of stream to write to
;       value  - low order word of this is value to write
;       padding - type of padding (0=none, 1=spaces, 2=zeroes)
;       padlen  - minimal length of number. ignored if padding=0.
; ret: CF set on error
; error: ERR_OUT_OF_RANGE - padlen > 4096
;        ERR_FASMLIB      - internal error in FASMLIB, please report
;        +extconv.d2dec
;        +stream.write
;=============================================================================
proc text.extwrite.w2dec stream, value, padding, padlen
	movzx	eax, word [value]
	libcall text.extwrite.d2dec, [stream], eax, [padding], [padlen]
	ret
endp



;=============================================================================
; text.extwrite.d2dec
; desc: writes unsigned dword (32 bit) to stream as decimal string.
; args: stream - handle of stream to write to
;       value  - value to write
;       padding - type of padding (0=none, 1=spaces, 2=zeroes)
;       padlen  - minimal length of number. ignored if padding=0.
; ret: CF set on error
; error: ERR_OUT_OF_RANGE - padlen > 4096
;        ERR_FASMLIB      - internal error in FASMLIB, please report
;        +extconv.d2dec
;        +stream.write
;=============================================================================
proc text.extwrite.d2dec stream, value, padding, padlen
local convlen dd ?

	push	ebx ecx

	;ECX = size of needed buffer, at least 16
	mov	ecx, [padlen]
	cmp	[padding], 0
	jne	@f
	xor	ecx, ecx
	@@:
	inc	ecx
	add	ecx, 3
	and	ecx, not 3  ;align to dword
	mov	eax, 16
	cmp	ecx, eax
	cmovb	ecx, eax

	;sanity-check for padlen > 4096, to prevent big allocation on stack
	cmp	ecx, 4096
	ja	.error_padlen_too_big

	;EBX = allocte padlen+1 buffer on stack
	sub	esp, ecx
	mov	ebx, esp

	;convert
	libcall extconv.d2dec, ebx, ecx, [value], [padding], [padlen], addr convlen
	jc	.error_convert

	;display
	libcall stream.write, [stream], ebx, [convlen]
	jc	.error_write

	;remove buffer from stack
	add	esp, ecx

.r:	pop	ebx
	ret
.rc:	stc
	jmp	.r

.error_write:
	add	esp, ecx
	jmp	.rc

.error_padlen_too_big:
	mov	eax, ERR_OUT_OF_RANGE
	jmp	.rc

.error_convert:
	add	esp, ecx
	;if returned error is ERR_BUFFER_FULL, it's FASMLIB bug
	cmp	eax, ERR_BUFFER_FULL
	jne	.rc
	mov	eax, ERR_FASMLIB
	jmp	.rc
endp



;=============================================================================
; text.extwrite.ib2hex
; desc: writes signed byte (8bit) to stream as zero-padded hex string
; args: stream - handle of stream to write to
;       value  - low order byte of this is value to write
;       padding - type of padding (0=none, 1=spaces, 2=zeroes)
;       padlen  - minimal length of number. ignored if padding=0.
;       plus    - 0=no "+", 1="+" for positive numbers
;                 2=" " for positive, only useful when padding=2
; ret: CF set on error
; error: ERR_OUT_OF_RANGE - padlen > 4096
;        ERR_FASMLIB      - internal error in FASMLIB, please report
;        +extconv.id2hex
;        +stream.write
;=============================================================================
proc text.extwrite.ib2hex stream, value, padding, padlen, plus
	movsx	eax, byte [value]
	libcall text.extwrite.id2hex, [stream], eax, [padding], [padlen], [plus]
	ret
endp



;=============================================================================
; text.extwrite.iw2hex
; desc: writes signed word (16bit) to stream as zero-padded hex string
; args: stream - handle of stream to write to
;       value  - low order word of this is value to write
;       padding - type of padding (0=none, 1=spaces, 2=zeroes)
;       padlen  - minimal length of number. ignored if padding=0.
;       plus    - 0=no "+", 1="+" for positive numbers
;                 2=" " for positive, only useful when padding=2
; ret: CF set on error
; error: ERR_OUT_OF_RANGE - padlen > 4096
;        ERR_FASMLIB      - internal error in FASMLIB, please report
;        +extconv.id2hex
;        +stream.write
;=============================================================================
proc text.extwrite.iw2hex stream, value, padding, padlen, plus
	movsx	eax, word [value]
	libcall text.extwrite.id2hex, [stream], eax, [padding], [padlen], [plus]
	ret
endp



;=============================================================================
; text.extwrite.id2hex
; desc: writes signed dword (32bit) to stream as zero-padded hex string
; args: stream - handle of stream to write to
;       value  - value to write
;       padding - type of padding (0=none, 1=spaces, 2=zeroes)
;       padlen  - minimal length of number. ignored if padding=0.
;       plus    - 0=no "+", 1="+" for positive numbers
;                 2=" " for positive, only useful when padding=2
; ret: CF set on error
; error: ERR_OUT_OF_RANGE - padlen > 4096
;        ERR_FASMLIB      - internal error in FASMLIB, please report
;        +extconv.id2hex
;        +stream.write
;=============================================================================
proc text.extwrite.id2hex stream, value, padding, padlen, plus
local convlen dd ?

	push	ebx ecx

	;ECX = size of needed buffer, at least 16
	mov	ecx, [padlen]
	inc	ecx
	inc	ecx
	add	ecx, 3
	and	ecx, not 3  ;align to dword
	mov	eax, 16
	cmp	ecx, eax
	cmovb	ecx, eax

	;sanity-check for padlen > 4096, to prevent big allocation on stack
	cmp	ecx, 4096
	ja	.error_padlen_too_big

	;EBX = allocte padlen+1 buffer on stack
	sub	esp, ecx
	mov	ebx, esp

	;convert
	libcall extconv.id2hex, ebx, ecx, [value], [padding], [padlen], [plus], addr convlen
	jc	.error_convert

	;display
	libcall stream.write, [stream], ebx, [convlen]
	jc	.error_write

	;remove buffer from stack
	add	esp, ecx

.r:	pop	ebx
	ret
.rc:	stc
	jmp	.r

.error_write:
	add	esp, ecx
	jmp	.rc

.error_padlen_too_big:
	mov	eax, ERR_OUT_OF_RANGE
	jmp	.rc

.error_convert:
	add	esp, ecx
	;if returned error is ERR_BUFFER_FULL, it's FASMLIB bug
	cmp	eax, ERR_BUFFER_FULL
	jne	.rc
	mov	eax, ERR_FASMLIB
	jmp	.rc

endp



;=============================================================================
; text.extwrite.ib2dec
; desc: writes signed byte (8 bit) to stream as decimal string
; args: stream - handle of stream to write to
;       value  - low order byte of this is value to write
;       padding - type of padding (0=none, 1=spaces, 2=zeroes)
;       padlen  - minimal length of number. ignored if padding=0.
;       plus    - 0=no "+", 1="+" for positive numbers
;                 2=" " for positive, only useful when padding=2
; ret: CF set on error
; error: ERR_OUT_OF_RANGE - padlen > 4096
;        ERR_FASMLIB      - internal error in FASMLIB, please report
;        +extconv.id2hex
;        +stream.write
;=============================================================================
proc text.extwrite.ib2dec stream, value, padding, padlen, plus
	movsx	eax, byte [value]
	libcall text.extwrite.id2dec, [stream], eax, [padding], [padlen], [plus]
	ret
endp



;=============================================================================
; text.extwrite.iw2dec
; desc: writes signed word (16 bit) to stream as decimal string
; args: stream - handle of stream to write to
;       value  - low order word of this is value to write
;       padding - type of padding (0=none, 1=spaces, 2=zeroes)
;       padlen  - minimal length of number. ignored if padding=0.
;       plus    - 0=no "+", 1="+" for positive numbers
;                 2=" " for positive, only useful when padding=2
; ret: CF set on error
; error: ERR_OUT_OF_RANGE - padlen > 4096
;        ERR_FASMLIB      - internal error in FASMLIB, please report
;        +extconv.id2hex
;        +stream.write
;=============================================================================
proc text.extwrite.iw2dec stream, value, padding, padlen, plus
	movsx	eax, word [value]
	libcall text.extwrite.id2dec, [stream], eax, [padding], [padlen], [plus]
	ret
endp



;=============================================================================
; text.extwrite.id2dec
; desc: writes signed dword (32 bit) to stream as decimal string.
; args: stream - handle of stream to write to
;       value  - value to write
;       padding - type of padding (0=none, 1=spaces, 2=zeroes)
;       padlen  - minimal length of number. ignored if padding=0.
;       plus    - 0=no "+", 1="+" for positive numbers
;                 2=" " for positive, only useful when padding=2
; ret: CF set on error
; error: ERR_OUT_OF_RANGE - padlen > 4096
;        ERR_FASMLIB      - internal error in FASMLIB, please report
;        +extconv.ib2hex
;        +stream.write
;=============================================================================
proc text.extwrite.id2dec stream, value, padding, padlen, plus
local convlen dd ?

	push	ebx ecx

	;ECX = size of needed buffer, at least 16
	mov	ecx, [padlen]
	inc	ecx
	inc	ecx
	add	ecx, 3
	and	ecx, not 3  ;align to dword
	mov	eax, 16
	cmp	ecx, eax
	cmovb	ecx, eax

	;sanity-check for padlen > 4096, to prevent big allocation on stack
	cmp	ecx, 4096
	ja	.error_padlen_too_big

	;EBX = allocte padlen+1 buffer on stack
	sub	esp, ecx
	mov	ebx, esp

	;convert
	libcall extconv.id2dec, ebx, ecx, [value], [padding], [padlen], [plus], addr convlen
	jc	.error_convert

	;display
	libcall stream.write, [stream], ebx, [convlen]
	jc	.error_write

	;remove buffer from stack
	add	esp, ecx

.r:	pop	ebx
	ret
.rc:	stc
	jmp	.r

.error_write:
	add	esp, ecx
	jmp	.rc

.error_padlen_too_big:
	mov	eax, ERR_OUT_OF_RANGE
	jmp	.rc

.error_convert:
	add	esp, ecx
	;if returned error is ERR_BUFFER_FULL, it's FASMLIB bug
	cmp	eax, ERR_BUFFER_FULL
	jne	.rc
	mov	eax, ERR_FASMLIB
	jmp	.rc
endp



;=============================================================================
; text.extread.chars
; desc: reads from the stream all chars that are in given list of characters.
; args: stream   - stream to skip characters from
;       buffer   - pointer to buffer, where the characters will be stored
;       buflen   - length of buffer
;       charlist - null-terminated string containing characters to skip
;       maxlen   - maximum number of characters to skip. ignored if 0
;       len      - number of characters readen. this is same value as one
;                  returned in EAX, this is only useful on ERR_BUFFER_FULL
;                  ignored if 0.
; ret: CF=1 on error, otherwise
;      EAX = number of characters skipped, and
;      ZF=1 if EAX=0 (no characters readen)
; note: - character 0 always stops skipping
; error: ERR_ZERO_SIZE   - buflen=0. len is set to 0
;        ERR_BUFFER_FULL - buffer is full. only buflen characters readen.
;                          len is set to buflen
;        ERR_HANDLE_EOF  - EOF reached without reading anything. len=0
;        ERR_FASMLIB     - bug in FASMLIB! Please report
;        +stream.peek
;        +stream.peeknb
;        +stream.skip
;        +str.poschar
;=============================================================================
proc text.extread.chars stream, buffer, buflen, charlist, maxlen, len
local peeklen dd ?
	push	ebx ecx edx edi
	pushf

	;error if length of buffer is 0
	cmp	[buflen], 0
	je	.error_buflen_zero

	;EDX = maximum number of chars allowed to skip (0 if not specified)
	mov	edx, -1
	mov	eax, [maxlen]
	cmp	eax, 0
	cmovne	edx, eax

	;ECX = number of chars skipped
	xor	ecx, ecx

	;EDI = destination buffer
	mov	edi, [buffer]

	;prebuffer 16 chars :]
	libcall stream.peeknb, [stream], 16, addr peeklen
	jc	.rc

	;if we reached EOF without reading any char, return ERR_BUFFER_FULL
	libcall stream.peekchar, [stream]
	jc	.rc
	jo	.error_eof
	jmp	.parse_char

.read_char:

	;EBX = peek 1 char
	libcall stream.peekchar, [stream]
	jc	.rc
	jo	.done

.parse_char:
	movzx	ebx, al

	;find out if current char is in charlist
	libcall str.poschar, [charlist], -1, ebx
	jc	.rc

	;if no, skipping is done
	jz	.done

	;store character
	mov	[edi], bl

	;skip char from stream
	libcall stream.skip, [stream], 1
	jc	.rc
	jo	.error_internal

	;another char done
	inc	ecx
	inc	edi

	;error if we filled entire buffer
	cmp	ecx, [buflen]
	je	.error_buffer_full
	ja	.error_internal

	;end loop if we skipped enough characters
	cmp	ecx, edx
	je	.done
	ja	.error_internal

	;continue loop
	jmp	.read_char

.done:
	;store ending 0 to buffer
	mov	byte [edi], 0

	;store number of readen chars to len
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	[eax], ecx
	@@:

	;return ECX
	mov	eax, ecx

	;if we skipped 0 characters, return ZF=1
	cmp	ecx, 0
	je	.rz

	;otherwise return ZF=0

.rnz:	and	byte [esp], not (1 shl 6) ;clear ZF
.rnc:	and	byte [esp], not 1 ;clear CF
.r:	popf
	pop	edi edx ecx ebx
	ret
.rc:	or	byte [esp], 1 ;set CF
	jmp	.r
.rz:	or	byte [esp], (1 shl 6) ;set ZF
	jmp	.rnc

.error_internal:
	mov	eax, ERR_FASMLIB
	jmp	.rc

.error_buffer_full:
	;store number of readen chars to len
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	[eax], ecx
	@@:

	mov	eax, ERR_BUFFER_FULL
	jmp	.rc

.error_buflen_zero:
	;store 0 to len
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	[eax], ecx
	@@:

	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_eof:
	mov	eax, ERR_HANDLE_EOF
	jmp	.rc

endp



;=============================================================================
; text.extread.until
; desc: reads from stream all chars until character from charlist is reached
; args: stream   - stream to read characters from
;       buffer   - pointer to buffer, where the characters will be stored
;       buflen   - length of buffer
;       charlist - null-terminated string containing characters on which to
;                  stop reading
;       maxlen   - maximum number of characters to read. ignored if 0
;       len      - number of characters readen. this is same value as one
;                  returned in EAX, this is only useful on ERR_BUFFER_FULL
;                  ignored if 0.
; ret: CF=1 on error, otherwise
;      EAX = number of characters readen, and
;      ZF=1 if EAX=0 (no characters readen)
; note: - character 0 always stops reading
; error: ERR_ZERO_SIZE   - buflen=0. len is set to 0
;        ERR_BUFFER_FULL - buffer is full. only buflen characters readen.
;                          len is set to buflen
;        ERR_HANDLE_EOF  - EOF reached without reading anything
;        ERR_FASMLIB     - bug in FASMLIB! Please report
;        +stream.peek
;        +stream.peeknb
;        +stream.skip
;        +str.poschar
;=============================================================================
proc text.extread.until stream, buffer, buflen, charlist, maxlen, len
local peeklen dd ?
	push	ebx ecx edx edi
	pushf

	;error if length of buffer is 0
	cmp	[buflen], 0
	je	.error_buflen_zero

	;EDX = maximum number of chars allowed to skip (0 if not specified)
	mov	edx, -1
	mov	eax, [maxlen]
	cmp	eax, 0
	cmovne	edx, eax

	;ECX = number of chars skipped
	xor	ecx, ecx

	;EDI = destination buffer
	mov	edi, [buffer]

	;prebuffer 16 chars :]
	libcall stream.peeknb, [stream], 16, addr peeklen
	jc	.rc

	;if we reached EOF without reading any char, return ERR_BUFFER_FULL
	libcall stream.peekchar, [stream]
	jc	.rc
	jo	.error_eof
	jmp	.parse_char

.read_char:

	;EBX = peek 1 char
	libcall stream.peekchar, [stream]
	jc	.rc
	jo	.done

.parse_char:
	movzx	ebx, al

	;find out if current char is in charlist
	libcall str.poschar, [charlist], -1, ebx
	jc	.rc

	;if yes, skipping is done
	jnz	.done

	;store character
	mov	[edi], bl

	;skip char from stream
	libcall stream.skip, [stream], 1
	jc	.rc
	jo	.error_internal

	;another char done
	inc	ecx
	inc	edi

	;error if we filled entire buffer
	cmp	ecx, [buflen]
	je	.error_buffer_full
	ja	.error_internal

	;end loop if we skipped enough characters
	cmp	ecx, edx
	je	.done
	ja	.error_internal

	;continue loop
	jmp	.read_char

.done:
	;store ending 0 to buffer
	mov	byte [edi], 0

	;store number of readen chars to len
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	[eax], ecx
	@@:

	;return ECX
	mov	eax, ecx

	;if we skipped 0 characters, return ZF=1
	cmp	ecx, 0
	je	.rz

	;otherwise return ZF=0

.rnz:	and	byte [esp], not (1 shl 6) ;clear ZF
.rnc:	and	byte [esp], not 1 ;clear CF
.r:	popf
	pop	edi edx ecx ebx
	ret
.rc:	or	byte [esp], 1 ;set CF
	jmp	.r
.rz:	or	byte [esp], (1 shl 6) ;set ZF
	jmp	.rnc

.error_internal:
	mov	eax, ERR_FASMLIB
	jmp	.rc

.error_buffer_full:
	;store number of readen chars to len
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	[eax], ecx
	@@:

	mov	eax, ERR_BUFFER_FULL
	jmp	.rc

.error_eof:
	mov	eax, ERR_HANDLE_EOF
	jmp	.rc

.error_buflen_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

endp



;=============================================================================
; desc: reads line from stream. Skips EOL to next line
; args: stream   - stream to read from
;       buffer   - pointer to buffer, where the characters will be stored
;       buflen   - length of buffer
;       charlist - null-terminated string containing characters on which to
;                  stop reading
;       maxlen   - maximum number of characters to skip. ignored if 0
;       len      - number of characters readen. this is same value as one
;                  returned in EAX, this is only useful on ERR_BUFFER_FULL
;                  ignored if 0.
; ret: CF=1 on error, otherwise
;      EAX = number of characters readen
;=============================================================================
proc text.extread.line, stream, buffer, buflen, maxlen, len
	push ebx

	;read the line into buffer
	libcall text.extread.until, [stream], [buffer], [buflen], text.eol_chars, [maxlen], [len]
	jc	.rc

	;done if we reached EOF
	jo	.r

	mov	ebx, eax

	;being smart-ass, we can skip character 13 if it's there
	libcall text.fetch.char, [stream], 13
	jc	.rc

	;then let's fetch character 10
	libcall text.fetch.char, [stream], 10
	jc	.rc

	mov	eax, ebx

	clc
.r:	pop	ebx
	ret
.rc:	stc
	jmp	.r

endp



;=============================================================================
; text.extread.hex2b
; desc: reading of hex byte (8 bit). extended version of text.read.hex2b.
;       reads hex digits (0-9, a-z, A-Z) and converts them to number, until it
;       reaches non-hex character, or until resulting number overflows.
; args: stream - handle of stream from which to read
;       maxlen - maximum number of characters to read. ignored if 0
;       skipbl - if 1, all leading blank characters are skipped. if you are
;                unsure, set to this 1
;       len    - pointer to variable, which on return, it holds number of
;                character read. ignored if 0
; ret: CF set on error, otherwise
;      EAX = value
; note: - if stream doesn't have dynamic buffer, then static buffer must be
;         big enough to hold entire number, or you get ERR_BUFFER_FULL
; error: ERR_HANDLE_EOF      - file pointer is on EOF. len contains 0
;        ERR_OUT_OF_RANGE    - value is > 0FFh
;                              in this case nothing is read from stream.
;                              len contains number of chars until overflow
;        ERR_NOT_HEXADECIMAL - first char isn't 0-9, a-z, A-Z
;                              in this case, nothing is read from stream.
;                              len contains 0
;        ERR_ZERO_VALUE      - variable pointed by len contains zero
;        ERR_FASMLIB         - bug in FASMLIB! please report it
;        +text.skip.chars
;        +stream.peeknb
;        +stream.peek
;        +stream.skip
;=============================================================================
proc text.extread.hex2b stream, maxlen, skipbl, len
	libcall text.extread.hex2d_size, [stream], 0x000000F0, [maxlen], [len], [skipbl]
	ret
endp



;=============================================================================
; text.extread.hex2w
; desc: reading of hex word (16bit). extended version of text.read.hex2w.
;       reads hex digits (0-9, a-z, A-Z) and converts them to number, until it
;       reaches non-hex character, or until resulting number overflows.
; args: stream - handle of stream from which to read
;       maxlen - maximum number of characters to read. ignored if 0
;       skipbl - if 1, all leading blank characters are skipped. if you are
;                unsure, set to this 1
;       len    - pointer to variable, which on return, it holds number of
;                character read. ignored if 0
; ret: CF set on error, otherwise
;      EAX = value
; note: - if stream doesn't have dynamic buffer, then static buffer must be
;         big enough to hold entire number, or you get ERR_BUFFER_FULL
; error: ERR_HANDLE_EOF      - file pointer is on EOF. len contains 0
;        ERR_OUT_OF_RANGE    - value is > 0FFFFh
;                              in this case nothing is read from stream.
;                              len contains number of chars until overflow
;        ERR_NOT_HEXADECIMAL - first char isn't 0-9, a-z, A-Z
;                              in this case, nothing is read from stream.
;                              len contains 0
;        ERR_ZERO_VALUE      - variable pointed by len contains zero
;        ERR_FASMLIB         - bug in FASMLIB! please report it
;        +text.skip.chars
;        +stream.peeknb
;        +stream.peek
;        +stream.skip
;=============================================================================
proc text.extread.hex2w stream, maxlen, skipbl, len
	libcall text.extread.hex2d_size, [stream], 0x0000F000, [maxlen], [len], [skipbl]
	ret
endp



;=============================================================================
; text.extread.hex2d
; desc: reading of hex dword (32bit). extended version of text.read.hex2d.
;       reads hex digits (0-9, a-z, A-Z) and converts them to number, until it
;       reaches non-hex character, or until resulting number overflows.
; args: stream - handle of stream from which to read
;       maxlen - maximum number of characters to read. ignored if 0
;       skipbl - if 1, all leading blank characters are skipped. if you are
;                unsure, set to this 1
;       len    - pointer to variable, which on return, it holds number of
;                character read. ignored if 0
; ret: CF set on error, otherwise
;      EAX = value
; note: - if stream doesn't have dynamic buffer, then static buffer must be
;         big enough to hold entire number, or you get ERR_BUFFER_FULL
; error: ERR_HANDLE_EOF      - file pointer is on EOF. len contains 0
;        ERR_OUT_OF_RANGE    - value is > 0FFFFFFFFh
;                              in this case nothing is read from stream.
;                              len contains number of chars until overflow
;        ERR_NOT_HEXADECIMAL - first char isn't 0-9, a-z, A-Z
;                              in this case, nothing is read from stream.
;                              len contains 0
;        ERR_ZERO_VALUE      - variable pointed by len contains zero
;        ERR_FASMLIB         - bug in FASMLIB! please report it
;        +text.skip.chars
;        +stream.peeknb
;        +stream.peek
;        +stream.skip
;=============================================================================
proc text.extread.hex2d stream, maxlen, skipbl, len
	libcall text.extread.hex2d_size, [stream], 0xF0000000, [maxlen], [len], [skipbl]
	ret
endp
if used text.extread.hex
  label text.extread.hex at text.extread.hex2d
end if



;=============================================================================
; text.extread.dec2b
; desc: reading of decimal byte (8 bit). extended version of text.read.dec2b.
;       reads decimal digits (0-9) and converts them to number, until it
;       reaches non-digit, or until resulting number overflows.
; args: stream - handle of stream from which to read
;       maxlen - maximum number of characters to read
;       skipbl - if 1, all leading blank characters are skipped. if you are
;                unsure, set to this 1
;       len    - pointer to variable, which on return, it holds number of
;                character read. ignored if 0
; ret: CF set on error, otherwise
;      EAX = value
; note: - if stream doesn't have dynamic buffer, then static buffer must be
;         big enough to hold entire number, or you get ERR_BUFFER_FULL
; error: ERR_HANDLE_EOF      - file pointer is on EOF. len contains 0
;        ERR_OUT_OF_RANGE    - value is > 255
;                              in this case nothing is read from stream.
;                              len contains number of chars until overflow
;        ERR_NOT_HEXADECIMAL - first char isn't decimal digit (0-9)
;                              in this case, nothing is read from stream.
;                              len contains 0
;        ERR_ZERO_VALUE      - variable pointed by len contains zero
;        ERR_FASMLIB         - bug in FASMLIB! please report it
;        +text.skip.chars
;        +stream.peeknb
;        +stream.peek
;        +stream.skip
;=============================================================================
proc text.extread.dec2b stream, maxlen, skipbl, len
	libcall text.extread.dec2d_size, [stream], 255, [maxlen], [len], [skipbl]
	ret
endp



;=============================================================================
; text.extread.dec2w
; desc: reading of decimal word (16bit). extended version of text.read.dec2b.
;       reads decimal digits (0-9) and converts them to number, until it
;       reaches non-digit, or until resulting number overflows.
; args: stream - handle of stream from which to read
;       maxlen - maximum number of characters to read
;       skipbl - if 1, all leading blank characters are skipped. if you are
;                unsure, set to this 1
;       len    - pointer to variable, which on return, it holds number of
;                character read. ignored if 0
; ret: CF set on error, otherwise
;      EAX = value
; note: - if stream doesn't have dynamic buffer, then static buffer must be
;         big enough to hold entire number, or you get ERR_BUFFER_FULL
; error: ERR_HANDLE_EOF      - file pointer is on EOF. len contains 0
;        ERR_OUT_OF_RANGE    - value is > 65535
;                              in this case nothing is read from stream.
;                              len contains number of chars until overflow
;        ERR_NOT_HEXADECIMAL - first char isn't decimal digit (0-9)
;                              in this case, nothing is read from stream.
;                              len contains 0
;        ERR_ZERO_VALUE      - variable pointed by len contains zero
;        ERR_FASMLIB         - bug in FASMLIB! please report it
;        +text.skip.chars
;        +stream.peeknb
;        +stream.peek
;        +stream.skip
;=============================================================================
proc text.extread.dec2w stream, maxlen, skipbl, len
	libcall text.extread.dec2d_size, [stream], 65535, [maxlen], [len], [skipbl]
	ret
endp



;=============================================================================
; text.extread.dec2d
; desc: reading of decimal dword (32bit). extended version of text.read.dec2b.
;       reads decimal digits (0-9) and converts them to number, until it
;       reaches non-digit, or until resulting number overflows.
; args: stream - handle of stream from which to read
;       maxlen - maximum number of characters to read
;       skipbl - if 1, all leading blank characters are skipped. if you are
;                unsure, set to this 1
;       len    - pointer to variable, which on return, it holds number of
;                character read. ignored if 0
; ret: CF set on error, otherwise
;      EAX = value
; note: - if stream doesn't have dynamic buffer, then static buffer must be
;         big enough to hold entire number, or you get ERR_BUFFER_FULL
; error: ERR_HANDLE_EOF      - file pointer is on EOF. len contains 0
;        ERR_OUT_OF_RANGE    - value is > 4294967295
;                              in this case nothing is read from stream.
;                              len contains number of chars until overflow
;        ERR_NOT_HEXADECIMAL - first char isn't decimal digit (0-9)
;                              in this case, nothing is read from stream.
;                              len contains 0
;        ERR_ZERO_VALUE      - variable pointed by len contains zero
;        ERR_FASMLIB         - bug in FASMLIB! please report it
;        +text.skip.chars
;        +stream.peeknb
;        +stream.peek
;        +stream.skip
;=============================================================================
proc text.extread.dec2d stream, maxlen, skipbl, len
	libcall text.extread.dec2d_size, [stream], -1, [maxlen], [len], [skipbl]
	ret
endp
if used text.extread.dec
  label text.extread.dec at text.extread.dec2d
end if



;=============================================================================
; text.extread.hex2d_size
; desc: INTERNAL! does job hex2b, hex2w, hex2d
;=============================================================================
proc text.extread.hex2d_size stream, mask, maxlen, skipbl, len
local buflen dd ?
	push	ebx ecx edx esi

	;skip blank chars
	;smartass thingy, but without it it would confuse people
	cmp	[skipbl], 1
	jne	@f
	libcall text.skip.chars, [stream], text.blank_chars
	jc	.rc
	@@:

	;EDX = maximum number of chars allowed to read  (0 if not specified)
	mov	edx, -1
	mov	eax, [maxlen]
	cmp	eax, 0
	cmovne	edx, eax

	;error if maximal number of chars to read is zero
	cmp	edx, 0
	je	.error_len_zero

	;peek at first char
	libcall stream.peekchar, [stream]
	jc	.rc
	jo	.error_eof

	;error if it isn't hex char
	cmp	al, '0'
	jb	.error_not_numeric
	cmp	al, '9'
	jbe	@f
	cmp	al, 'A'
	jb	.error_not_numeric
	cmp	al, 'F'
	jbe	@f
	cmp	al, 'a'
	jb	.error_not_numeric
	cmp	al, 'f'
	ja	.error_not_numeric
	@@:

	;ECX = number of char converted
	xor	ecx, ecx

	;EBX = result
	xor	ebx, ebx

.load_char:

	;non-blocking peek 10 more characters
	lea	eax, [ecx+10]
	libcall stream.peeknb, [stream], eax, addr buflen
	jc	.rc
	mov	esi, eax

	;if we got at least one more, go convert it
	cmp	[buflen], ecx
	ja	.convert_char

	;peek 1 more char
	lea	eax, [ecx+1]
	libcall stream.peek, [stream], eax, addr buflen
	jc	.rc
	jo	.done
	mov	esi, eax

.convert_char:
	mov	al, [esi + ecx]

	;end if it isn't hex digit
	cmp	al, '0'
	jb	.done
	cmp	al, '9'
	jbe	.numeric
	cmp	al, 'A'
	jb	.done
	cmp	al, 'F'
	jbe	.upcase_letter
	cmp	al, 'a'
	jb	.done
	cmp	al, 'f'
	ja	.done

.lowcase_letter:
	sub	al, 'a'-10
	jmp	.add_digit

.upcase_letter:
	sub	al, 'A'-10
	jmp	.add_digit

.numeric:
	sub	al, '0'

.add_digit:
	;check overflow
	test	ebx, [mask]
	jnz	.error_overflow

	;add digit
	shl	ebx,4
	movzx	eax, al
	add	ebx, eax

	;check if we have read enough chars
	inc	ecx
	cmp	ecx, edx
	ja	.error_internal
	je	.done

	;if we don't have another char buffered, read it. otherwise continue converting next char
	cmp	ecx, [buflen]
	ja	.error_internal
	jb	.convert_char
	jmp	.load_char

.done:
	;skip readen chars
	mov	eax, [buflen]
	libcall stream.skip, [stream], eax
	jc	.rc
	jo	.error_internal

	;save number of chars readen to variable pointet by len, if desired
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	[eax], ecx
	@@:

	;store result
	mov	eax, ebx

	clc
.r:	pop	esi edx ecx ebx
	ret
.rc:	stc
	jmp	.r


.error_eof:
	;save 0 to variable pointet by len, if desired
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], 0
	@@:

	mov	eax, ERR_HANDLE_EOF
	jmp	.rc

.error_not_numeric:
	;save 0 to variable pointet by len, if desired
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], 0
	@@:

	mov	eax, ERR_NOT_HEXADECIMAL
	jmp	.rc

.error_internal:
	mov	eax, ERR_FASMLIB
	jmp	.rc

.error_len_zero:
	mov	eax, ERR_ZERO_VALUE
	jmp	.rc

.error_overflow:
	;save ecx to variable pointet by len, if desired
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	[eax], ecx
	@@:

	mov	eax, ERR_OUT_OF_RANGE
	jmp	.rc
endp



;=============================================================================
; text.extread.dec2d_size
; desc: INTERNAL! does job dec2b, dec2w, dec2d
;=============================================================================
proc text.extread.dec2d_size stream, size, maxlen, skipbl, len
local buflen dd ?
	push	ebx ecx edx esi

	;skip blank chars
	;smartass thingy, but without it it would confuse people
	cmp	[skipbl], 1
	jne	@f
	libcall text.skip.chars, [stream], text.blank_chars
	jc	.rc
	@@:

	;EDX = maximum number of chars allowed to read  (0 if not specified)
	mov	edx, -1
	mov	eax, [maxlen]
	cmp	eax, 0
	cmovne	edx, eax

	;peek at first char
	libcall stream.peekchar, [stream]
	jc	.rc
	jo	.error_eof

	;error if it isn't hex char
	cmp	al, '0'
	jb	.error_not_numeric
	cmp	al, '9'
	ja	.error_not_numeric

	;ECX = number of char converted
	xor	ecx, ecx

	;EBX = result
	xor	ebx, ebx

.load_char:

	;non-blocking peek 10 more characters
	lea	eax, [ecx+10]
	libcall stream.peeknb, [stream], eax, addr buflen
	jc	.rc
	mov	esi, eax

	;if we got at least one more, go convert it
	cmp	[buflen], ecx
	ja	.convert_char

	;peek 1 more char
	lea	eax, [ecx+1]
	libcall stream.peek, [stream], eax, addr buflen
	jc	.rc
	jo	.done
	mov	esi, eax

.convert_char:
	movzx	eax, byte [esi + ecx]

	;end if it isn't hex digit
	cmp	al, '0'
	jb	.done
	cmp	al, '9'
	ja	.done

	;convert to hex digit
	sub	al, '0'

	;miltply value by 10, and add converted digit to it
	push	edx eax esi
	xor	edx, edx
	mov	esi, 10
	mov	eax, ebx
	mul	esi
	pop	esi
	mov	ebx, eax
	pop	eax
	add	ebx, eax
	mov	eax, edx
	pop	edx

	;check for overflow
	jc	.error_overflow
	cmp	eax, 0
	ja	.error_overflow
	cmp	ebx, [size]
	ja	.error_overflow

	;check if we have read enough chars
	inc	ecx
	cmp	ecx, edx
	ja	.error_internal
	je	.done

	;if we don't have another char buffered, read it. otherwise continue converting next char
	cmp	ecx, [buflen]
	ja	.error_internal
	jb	.convert_char
	jmp	.load_char

.done:
	;skip readen chars
	libcall stream.skip, [stream], ecx
	jc	.rc
	jo	.error_internal

	;save number of chars readen to variable pointet by len, if desired
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	[eax], ecx
	@@:

	;store result
	mov	eax, ebx

	clc
.r:	pop	esi edx ecx ebx
	ret
.rc:	stc
	jmp	.r


.error_eof:
	;save 0 to variable pointet by len, if desired
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], 0
	@@:

	mov	eax, ERR_HANDLE_EOF
	jmp	.rc

.error_not_numeric:
	;save 0 to variable pointet by len, if desired
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], 0
	@@:

	mov	eax, ERR_NOT_DECIMAL
	jmp	.rc

.error_internal:
	mov	eax, ERR_FASMLIB
	jmp	.rc

.error_len_zero:
	;save 0 to variable pointet by len
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], 0
	@@:

	mov	eax, ERR_ZERO_VALUE
	jmp	.rc

.error_overflow:
	;save ecx to variable pointet by len
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	[eax], ecx
	@@:

	mov	eax, ERR_OUT_OF_RANGE
	jmp	.rc
endp



;=============================================================================
; text.extskip.chars
; desc: reads from the stream all chars that are in given list of characters.
; args: stream   - stream to skip characters from
;       charlist - null-terminated string containing characters to skip.
;       maxlen   - maximum number of characters to skip
; ret: CF=1 on error, otherwise
;      OF=1 if EOF was reached, and
;      EAX = number of characters skipped
; note: - character 0 always stops skipping
; error: ERR_FASMLIB - bug in FASMLIB! Please report
;        +stream.peek
;        +stream.peeknb
;        +stream.skip
;        +str.poschar
;=============================================================================
proc text.extskip.chars stream, charlist, maxlen
local buflen dd ?
	push	ecx edx esi
	pushf

	;EDX = maximum number of chars allowed to skip (0 if not specified)
	mov	edx, -1
	mov	eax, [maxlen]
	cmp	eax, 0
	cmovne	edx, eax

	;ECX = number of chars skipped
	xor	ecx, ecx

	;prebuffer 16 chars :]    TODO
	libcall stream.peeknb, [stream], 16, addr buflen
	jc	.rc

	;check if we are on EOF
	libcall stream.peekchar, [stream]
	jc	.rc
	jo	.eof

.skip_char:

	;peek 1 char
	libcall stream.peekchar, [stream]
	jc	.rc
	jo	.done

	;find out if current char is in charlist
	libcall str.poschar, [charlist], -1, eax
	jc	.rc

	;if no, skipping is done
	jz	.done

	;skip char
	libcall stream.skip, [stream], 1
	jc	.rc
	jo	.error_internal

	;end loop if we skipped enough characters
	inc	ecx
	cmp	ecx, edx
	je	.done
	ja	.error_internal

	;continue loop
	jmp	.skip_char

.eof:
	;return ECX, OF=1
	mov	eax, ecx
	jmp	.ro

.done:

	;return ECX, OF=0
	mov	eax, ecx

.rno:	and	word [esp], not (1 shl 11) ;clear OF
.rnc:	and	byte [esp], not 1 ;clear CF
.r:	popf
	pop	esi edx ecx
	ret
.rc:	or	byte [esp], 1 ;set CF
	jmp	.r
.ro:	or	word [esp], (1 shl 11) ;set OF
	jmp	.rnc

.error_internal:
	mov	eax, ERR_FASMLIB
	jmp	.rc

endp



;=============================================================================
; text.extskip.until
; desc: skips from the stream all chars until some character from charlist is
;       reached
; args: stream   - stream to skip characters from
;       charlist - null-terminated string containing characters not to skip
;       maxlen   - maximum number of characters to skip. ignored if 0.
; ret: CF=1 on error, otherwise
;      OF=1 if we are on EOF, otherwise
;      EAX = number of characters skipped, and
; note: - character 0 always stops skipping
; error: ERR_FASMLIB - bug in FASMLIB! Please report
;        +stream.peek
;        +stream.peeknb
;        +stream.skip
;        +str.poschar
;=============================================================================
proc text.extskip.until stream, charlist, maxlen
local buflen dd ?
	push	ecx edx esi
	pushf

	;EDX = maximum number of chars allowed to skip (0 if not specified)
	mov	edx, -1
	mov	eax, [maxlen]
	cmp	eax, 0
	cmovne	edx, eax

	;ECX = number of chars skipped
	xor	ecx, ecx

	;prebuffer 16 chars :]
	libcall stream.peeknb, [stream], 16, addr buflen
	jc	.rc

	;check if we are on EOF
	libcall stream.peekchar, [stream]
	jc	.rc
	jo	.eof

.skip_char:

	;EAX = peek 1 char
	libcall stream.peek, [stream], 1, 0
	jc	.rc
	jo	.done
	movzx	eax, byte [eax]

	;done if it is char 0
	cmp	al, 0
	je	.done

	;find out if current char is in charlist
	libcall str.poschar, [charlist], -1, eax
	jc	.rc

	;if yes, skipping is done
	jnz	.done

	;skip char
	libcall stream.skip, [stream], 1
	jc	.rc
	jo	.error_internal

	;end loop if we skipped enough characters
	inc	ecx
	cmp	ecx, edx
	je	.done
	ja	.error_internal

	;continue loop
	jmp	.skip_char

.eof:
	;return ECX, OF=1
	mov	eax, ecx
	jmp	.ro

.done:

	;return ECX, OF=0
	mov	eax, ecx

.rno:	and	word [esp], not (1 shl 11) ;clear OF
.rnc:	and	byte [esp], not 1 ;clear CF
.r:	popf
	pop	esi edx ecx
	ret
.rc:	or	byte [esp], 1 ;set CF
	jmp	.r
.ro:	or	word [esp], (1 shl 11) ;set OF
	jmp	.rnc

.error_internal:
	mov	eax, ERR_FASMLIB
	jmp	.rc

endp



;=============================================================================
; text.extskip.hex
; desc: reads hexadecimal number (character 0-9, a-z, A-Z) off the stream
; args: stream   - stream to skip number from
;       maxlen   - maximum number of characters to skip. ignored if 0.
; ret: CF=1 on error, otherwise
;      OF=1 if we are on EOF, otherwise
;      EAX = number of characters skipped, and
; note: - this skips without error long numbers, that would cause
;         ERR_OUT_OF_RANGE with text.read or text.fetch
; error: +text.extskip.chars
;=============================================================================
proc text.extskip.hex stream, maxlen
	libcall text.extskip.chars, stream.stdin, text.hex_chars, [maxlen]
	ret
endp



;=============================================================================
; text.extskip.dec
; desc: skips decimal number (character 0-9) off the stream
; args: stream   - stream to skip number from
;       maxlen   - maximum number of characters to skip
; ret: CF=1 on error, otherwise
;      OF=1 if we are on EOF, otherwise
;      EAX = number of characters skipped, and
; note: - this skips without error long numbers, that would cause
;         ERR_OUT_OF_RANGE with text.read or text.fetch
; error: +text.extskip.chars
;=============================================================================
proc text.extskip.dec stream, maxlen
	libcall text.extskip.chars, stream.stdin, text.dec_chars, [maxlen]
	ret
endp



;by vid